#
# Copyright (c) 2016 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import atexit
import functools
import wrapt
import queue
import traceback
from threading import Thread, Lock, Event

from blatann.nrf.nrf_events import *
from blatann.nrf.nrf_types import *
from blatann.nrf.nrf_dll_load import driver
from pc_ble_driver_py.exceptions import NordicSemiException
import blatann.nrf.nrf_driver_types as util
from blatann.nrf.nrf_types.config import BleEnableConfig, BleConnConfig

logger = logging.getLogger(__name__)

NoneType = type(None)


# TODO: Do we really want to raise exceptions all the time?
def NordicSemiErrorCheck(wrapped=None, expected=driver.NRF_SUCCESS):
    if wrapped is None:
        return functools.partial(NordicSemiErrorCheck, expected=expected)

    @wrapt.decorator
    def wrapper(wrapped, instance, args, kwargs):
        logger.debug("{}{}".format(wrapped.__name__, args))
        result = wrapped(*args, **kwargs)
        if isinstance(result, (list, tuple)):
            err_code = result[0]
            result = result[1:]
            if len(result) == 1:
                result = result[0]
        else:
            err_code = result
            result = None

        if err_code != expected:
            try:
                err_string = 'Error code: {}'.format(NrfError(err_code))
            except ValueError:
                err_string = 'Error code: 0x{:04x}, {}'.format(err_code, err_code)
            raise NordicSemiException('Failed to {}. {}'.format(wrapped.__name__, err_string))
        return result

    return wrapper(wrapped)


class NrfDriverObserver(object):
    def on_driver_event(self, nrf_driver, event):
        pass


class NrfDriver(object):
    api_lock = Lock()
    default_baud_rate = 1000000
    ATT_MTU_DEFAULT = driver.BLE_GATT_ATT_MTU_DEFAULT

    def __init__(self, serial_port, baud_rate=None, log_driver_comms=False):
        if baud_rate is None:
            baud_rate = self.default_baud_rate

        self._events = queue.Queue()
        self._event_thread = None
        self._event_loop = False
        self._event_stopped = Event()
        self.observers = []
        self.ble_enable_params = None
        self._event_observers = {}
        self._event_observer_lock = Lock()
        self._log_driver_comms = log_driver_comms
        self._serial_port = serial_port

        phy_layer = driver.sd_rpc_physical_layer_create_uart(serial_port,
                                                             baud_rate,
                                                             driver.SD_RPC_FLOW_CONTROL_NONE,
                                                             driver.SD_RPC_PARITY_NONE)
        link_layer = driver.sd_rpc_data_link_layer_create_bt_three_wire(phy_layer, 100)
        transport_layer = driver.sd_rpc_transport_layer_create(link_layer, 100)
        self.rpc_adapter = driver.sd_rpc_adapter_create(transport_layer)

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def open(self):
        if self.is_open:
            logger.warning("Trying to open already opened driver")
            return driver.NRF_SUCCESS

        err_code = driver.sd_rpc_open(self.rpc_adapter,
                                      self._status_handler,
                                      self.ble_evt_handler,
                                      self._log_message_handler)

        if err_code == driver.NRF_SUCCESS:
            self._event_thread = Thread(target=self._event_handler, name="{}_Event".format(self._serial_port))
            # Note: We create a daemon thread and then register an exit handler
            #       to make sure this thread stops. This ensures that scripts that
            #       stop because of ctrl-c interrupt, compile errors or other problems
            #       do not keep hanging in the console, waiting for an infinite thread
            #       loop.
            atexit.register(self._event_thread_join)
            self._event_thread.daemon = True
            self._event_thread.start()
        return err_code

    def _event_thread_join(self):
        if self._event_thread is None:
            return
        self._event_loop = False
        self._event_stopped.wait(1)
        self._event_thread = None

    @property
    def is_open(self):
        return self._event_thread is not None

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def close(self):
        if not self.is_open:
            return driver.NRF_SUCCESS
        driver.sd_rpc_conn_reset(self.rpc_adapter, 0)
        retval = driver.sd_rpc_close(self.rpc_adapter)
        driver.sd_rpc_adapter_delete(self.rpc_adapter)
        self._event_thread_join()
        return retval

    def event_subscribe(self, handler, *event_types):
        for event_type in event_types:
            if not issubclass(event_type, BLEEvent):
                raise ValueError("Event type must be a valid BLEEvent class type. Got {}".format(event_type))
        with self._event_observer_lock:
            for event_type in event_types:
                # If event type not already in dict, create an empty list
                if event_type not in self._event_observers.keys():
                    self._event_observers[event_type] = []
                handlers = self._event_observers[event_type]
                if handler not in handlers:
                    handlers.append(handler)

    def event_unsubscribe(self, handler, *event_types):
        if not event_types:
            self.event_unsubscribe_all(handler)
            return

        with self._event_observer_lock:
            for event_type in event_types:
                handlers = self._event_observers.get(event_type, [])
                if handler in handlers:
                    handlers.remove(handler)

    def event_unsubscribe_all(self, handler):
        with self._event_observer_lock:
            for event_type, handlers in self._event_observers.items():
                if handler in handlers:
                    handlers.remove(handler)

    def observer_register(self, observer):
        with self._event_observer_lock:
            if observer not in self.observers:
                self.observers.append(observer)

    def observer_unregister(self, observer):
        with self._event_observer_lock:
            if observer in self.observers:
                self.observers.remove(observer)

    def ble_enable_params_setup(self):
        return BleEnableConfig()

    def adv_params_setup(self):
        return BLEGapAdvParams(interval_ms=40, timeout_s=180)

    def scan_params_setup(self):
        return BLEGapScanParams(interval_ms=200, window_ms=150, timeout_s=10)

    def conn_params_setup(self):
        return BLEGapConnParams(min_conn_interval_ms=15, max_conn_interval_ms=30,
                                conn_sup_timeout_ms=4000, slave_latency=0)

    def security_params_setup(self):
        return BLEGapSecParams(bond=True, mitm=True, le_sec_pairing=False, keypress_noti=False,
                               io_caps=BLEGapIoCaps.NONE, oob=False, min_key_size=8,max_key_size=16,
                               kdist_own=BLEGapSecKeyDist(), kdist_peer=BLEGapSecKeyDist(enc_key=True))

    """
    BLE Generic methods
    """
    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_conn_configure(self, conn_params):
        assert isinstance(conn_params, BleConnConfig)
        for tag, cfg in conn_params.get_configs():
            err = driver.sd_ble_cfg_set(self.rpc_adapter, tag, cfg, 0)
            if err != driver.NRF_SUCCESS:
                return err
        return driver.NRF_SUCCESS

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_enable(self, ble_enable_params=None):
        if not ble_enable_params:
            ble_enable_params = self.ble_enable_params_setup()
        assert isinstance(ble_enable_params, BleEnableConfig), 'Invalid argument type'
        self.ble_enable_params = ble_enable_params
        for tag, cfg in self.ble_enable_params.get_configs():
            err = driver.sd_ble_cfg_set(self.rpc_adapter, tag, cfg, 0)
            if err != driver.NRF_SUCCESS:
                return err

        return driver.sd_ble_enable(self.rpc_adapter, None)

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_user_mem_reply(self, conn_handle):
        return driver.sd_ble_user_mem_reply(self.rpc_adapter, conn_handle, None)

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_vs_uuid_add(self, uuid_base):
        assert isinstance(uuid_base, BLEUUIDBase), 'Invalid argument type'
        uuid_type = driver.new_uint8()

        err_code = driver.sd_ble_uuid_vs_add(self.rpc_adapter,
                                             uuid_base.to_c(),
                                             uuid_type)
        if err_code == driver.NRF_SUCCESS:
            uuid_base.type = driver.uint8_value(uuid_type)
        return err_code

    """
    GAP Methods
    """

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gap_addr_get(self):
        addr = driver.ble_gap_addr_t()
        err_code = driver.sd_ble_gap_addr_get(self.rpc_adapter, addr)
        if err_code == driver.NRF_SUCCESS:
            addr = BLEGapAddr.from_c(addr)
        else:
            addr = None
        return err_code, addr

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gap_addr_set(self, address):
        assert isinstance(address, BLEGapAddr), "Invalid argument type"
        addr_c = address.to_c()
        return driver.sd_ble_gap_addr_set(self.rpc_adapter, addr_c)

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gap_adv_start(self, adv_params=None, conn_cfg_tag=0):
        if not adv_params:
            adv_params = self.adv_params_setup()
        assert isinstance(adv_params, BLEGapAdvParams), 'Invalid argument type'
        return driver.sd_ble_gap_adv_start(self.rpc_adapter, adv_params.to_c(), conn_cfg_tag)

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gap_conn_param_update(self, conn_handle, conn_params):
        assert isinstance(conn_params, (BLEGapConnParams, NoneType)), 'Invalid argument type'
        if conn_params:
            conn_params = conn_params.to_c()
        return driver.sd_ble_gap_conn_param_update(self.rpc_adapter, conn_handle, conn_params)

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gap_adv_stop(self):
        return driver.sd_ble_gap_adv_stop(self.rpc_adapter)

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gap_scan_start(self, scan_params=None):
        if not scan_params:
            scan_params = self.scan_params_setup()
        assert isinstance(scan_params, BLEGapScanParams), 'Invalid argument type'
        return driver.sd_ble_gap_scan_start(self.rpc_adapter, scan_params.to_c())

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gap_scan_stop(self):
        return driver.sd_ble_gap_scan_stop(self.rpc_adapter)

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gap_connect(self, address, scan_params=None, conn_params=None, conn_cfg_tag=0):
        assert isinstance(address, BLEGapAddr), 'Invalid argument type'

        if not scan_params:
            scan_params = self.scan_params_setup()
        assert isinstance(scan_params, BLEGapScanParams), 'Invalid argument type'

        if not conn_params:
            conn_params = self.conn_params_setup()
        assert isinstance(conn_params, BLEGapConnParams), 'Invalid argument type'

        return driver.sd_ble_gap_connect(self.rpc_adapter,
                                         address.to_c(),
                                         scan_params.to_c(),
                                         conn_params.to_c(),
                                         conn_cfg_tag)

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gap_disconnect(self, conn_handle, hci_status_code=BLEHci.remote_user_terminated_connection):
        assert isinstance(hci_status_code, BLEHci), 'Invalid argument type'
        return driver.sd_ble_gap_disconnect(self.rpc_adapter,
                                            conn_handle,
                                            hci_status_code.value)

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gap_adv_data_set(self, adv_data=BLEAdvData(), scan_data=BLEAdvData()):
        assert isinstance(adv_data, BLEAdvData), 'Invalid argument type'
        assert isinstance(scan_data, BLEAdvData), 'Invalid argument type'
        (adv_data_len, p_adv_data) = adv_data.to_c()
        (scan_data_len, p_scan_data) = scan_data.to_c()

        return driver.sd_ble_gap_adv_data_set(self.rpc_adapter,
                                              p_adv_data,
                                              adv_data_len,
                                              p_scan_data,
                                              scan_data_len)

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gap_data_length_update(self, conn_handle, params=None):
        assert isinstance(params, (BLEGapDataLengthParams, NoneType))
        if isinstance(params, BLEGapDataLengthParams):
            params = params.to_c()
        return driver.sd_ble_gap_data_length_update(self.rpc_adapter, conn_handle, params, None)

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gap_phy_update(self, conn_handle, tx_phy=BLEGapPhy.auto, rx_phy=BLEGapPhy.auto):
        params = BLEGapPhys(tx_phy, rx_phy)
        return driver.sd_ble_gap_phy_update(self.rpc_adapter, conn_handle, params.to_c())

    """
    SMP Methods
    """

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gap_authenticate(self, conn_handle, sec_params):
        assert isinstance(sec_params, (BLEGapSecParams, NoneType)), 'Invalid argument type'
        return driver.sd_ble_gap_authenticate(self.rpc_adapter,
                                              conn_handle,
                                              sec_params.to_c() if sec_params else None)

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gap_sec_params_reply(self, conn_handle, sec_status, sec_params, sec_keyset):
        assert isinstance(sec_status, BLEGapSecStatus), 'Invalid argument type'
        assert isinstance(sec_params, (BLEGapSecParams, NoneType)), 'Invalid argument type'
        assert isinstance(sec_keyset, BLEGapSecKeyset), 'Invalid argument type'

        return driver.sd_ble_gap_sec_params_reply(self.rpc_adapter,
                                                  conn_handle,
                                                  sec_status.value,
                                                  sec_params.to_c() if sec_params else None,
                                                  sec_keyset.to_c())

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gap_auth_key_reply(self, conn_handle, key_type, key):
        if key is not None:
            key_buf = util.list_to_uint8_array(key).cast()
        else:
            key_buf = None
        return driver.sd_ble_gap_auth_key_reply(self.rpc_adapter, conn_handle, key_type, key_buf)

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gap_sec_info_reply(self, conn_handle, enc_info=None, irk=None, sign_info=None):
        assert isinstance(enc_info, (BLEGapEncryptInfo, NoneType)), "Invalid argument type"
        assert isinstance(irk, (BLEGapIdKey, NoneType)), "Invalid argument type"
        assert isinstance(sign_info, (BLEGapSignKey, NoneType)), "Invalid argument type"
        if enc_info is not None:
            enc_info = enc_info.to_c()
        if irk is not None:
            irk = irk.to_c().id_info
        if sign_info is not None:
            sign_info = sign_info.to_c()

        return driver.sd_ble_gap_sec_info_reply(self.rpc_adapter, conn_handle, enc_info, irk, sign_info)

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gap_encrypt(self, conn_handle, ediv, rand, ltk, lesc, auth):
        # TODO: Clean up
        # assert isinstance(sec_params, (BLEGapSecParams, NoneType)), 'Invalid argument type'
        # assert isinstance(sec_keyset, BLEGapSecKeyset), 'Invalid argument type'
        # print 'ediv %r' % master_id.ediv
        # print 'rand %r' % util.uint8_array_to_list(master_id.rand, 8)
        # print 'ltk  %r' % util.uint8_array_to_list(enc_info.ltk, enc_info.ltk_len)
        # print 'len  %r' % enc_info.ltk_len
        # print 'lesc %r' % enc_info.lesc
        # print 'auth %r' % enc_info.auth

        rand_arr = util.list_to_uint8_array(rand)
        ltk_arr = util.list_to_uint8_array(ltk)
        master_id = driver.ble_gap_master_id_t()
        master_id.ediv = ediv
        master_id.rand = rand_arr.cast()
        enc_info = driver.ble_gap_enc_info_t()
        enc_info.ltk_len = len(ltk)
        enc_info.ltk = ltk_arr.cast()
        enc_info.lesc = lesc
        enc_info.auth = auth
        return driver.sd_ble_gap_encrypt(self.rpc_adapter, conn_handle, master_id, enc_info)

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gap_lesc_dhkey_reply(self, conn_handle, dh_key):
        assert isinstance(dh_key, BLEGapDhKey)

        key = dh_key.to_c()
        return driver.sd_ble_gap_lesc_dhkey_reply(self.rpc_adapter, conn_handle, key)

    """
    GATTS Methods
    """
    # TODO: sd_ble_gatts_include_add, sd_ble_gatts_descriptor_add, sd_ble_gatts_sys_attr_set/get,
    # sd_ble_gatts_initial_user_handle_get, sd_ble_gatts_attr_get

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gatts_service_add(self, service_type, uuid, service_handle):
        handle = driver.new_uint16()
        uuid_c = uuid.to_c()
        err_code = driver.sd_ble_gatts_service_add(self.rpc_adapter,
                                                   service_type,
                                                   uuid_c,
                                                   handle)
        if err_code == driver.NRF_SUCCESS:
            service_handle.handle = driver.uint16_value(handle)
        return err_code

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gatts_characteristic_add(self, service_handle, char_md, attr_char_value, char_handle):
        # TODO type assertions
        handle_params = driver.ble_gatts_char_handles_t()
        err_code = driver.sd_ble_gatts_characteristic_add(self.rpc_adapter,
                                                          service_handle,
                                                          char_md.to_c(),
                                                          attr_char_value.to_c(),
                                                          handle_params)
        if err_code == driver.NRF_SUCCESS:
            char_handle.value_handle = handle_params.value_handle
            char_handle.user_desc_handle = handle_params.user_desc_handle
            char_handle.cccd_handle = handle_params.cccd_handle
            char_handle.sccd_handle = handle_params.sccd_handle
        return err_code

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gatts_rw_authorize_reply(self, conn_handle, authorize_reply_params):
        assert isinstance(authorize_reply_params, BLEGattsRwAuthorizeReplyParams)
        return driver.sd_ble_gatts_rw_authorize_reply(self.rpc_adapter, conn_handle, authorize_reply_params.to_c())

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gatts_value_get(self, conn_handle, attribute_handle, gatts_value):
        assert isinstance(gatts_value, BLEGattsValue)
        value_params = gatts_value.to_c()
        value_params.len = 512  # Allow up to 512 bytes to be read
        err_code = driver.sd_ble_gatts_value_get(self.rpc_adapter, conn_handle, attribute_handle, value_params)

        if err_code == driver.NRF_SUCCESS:
            value_out = BLEGattsValue.from_c(value_params)
            gatts_value.offset = value_out.offset
            gatts_value.value = value_out.value
        return err_code

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gatts_value_set(self, conn_handle, attribute_handle, gatts_value):
        assert isinstance(gatts_value, BLEGattsValue)
        value_params = gatts_value.to_c()
        return driver.sd_ble_gatts_value_set(self.rpc_adapter, conn_handle, attribute_handle, value_params)

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gatts_hvx(self, conn_handle, hvx_params):
        assert isinstance(hvx_params, BLEGattsHvx)
        return driver.sd_ble_gatts_hvx(self.rpc_adapter, conn_handle, hvx_params.to_c())

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gatts_service_changed(self, conn_handle, start_handle, end_handle):
        return driver.sd_ble_gatts_service_changed(self.rpc_adapter, conn_handle, start_handle, end_handle)

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gatts_exchange_mtu_reply(self, conn_handle, server_mtu):
        return driver.sd_ble_gatts_exchange_mtu_reply(self.rpc_adapter, conn_handle, server_mtu)

    """
    GATTC Methods
    """

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gattc_write(self, conn_handle, write_params):
        assert isinstance(write_params, BLEGattcWriteParams), 'Invalid argument type'
        return driver.sd_ble_gattc_write(self.rpc_adapter,
                                         conn_handle,
                                         write_params.to_c())

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gattc_prim_srvc_disc(self, conn_handle, srvc_uuid, start_handle):
        assert isinstance(srvc_uuid, (BLEUUID, NoneType)), 'Invalid argument type'
        return driver.sd_ble_gattc_primary_services_discover(self.rpc_adapter,
                                                             conn_handle,
                                                             start_handle,
                                                             srvc_uuid.to_c() if srvc_uuid else None)

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gattc_char_disc(self, conn_handle, start_handle, end_handle):
        handle_range = driver.ble_gattc_handle_range_t()
        handle_range.start_handle = start_handle
        handle_range.end_handle = end_handle
        return driver.sd_ble_gattc_characteristics_discover(self.rpc_adapter,
                                                            conn_handle,
                                                            handle_range)

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gattc_desc_disc(self, conn_handle, start_handle, end_handle):
        handle_range = driver.ble_gattc_handle_range_t()
        handle_range.start_handle = start_handle
        handle_range.end_handle = end_handle
        return driver.sd_ble_gattc_descriptors_discover(self.rpc_adapter,
                                                        conn_handle,
                                                        handle_range)

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gattc_attr_info_disc(self, conn_handle, start_handle, end_handle):
        handle_range = driver.ble_gattc_handle_range_t()
        handle_range.start_handle = start_handle
        handle_range.end_handle = end_handle
        return driver.sd_ble_gattc_attr_info_discover(self.rpc_adapter, conn_handle, handle_range)

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gattc_read(self, conn_handle, read_handle, offset=0):
        return driver.sd_ble_gattc_read(self.rpc_adapter, conn_handle, read_handle, offset)

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gattc_exchange_mtu_req(self, conn_handle, att_mtu_size):
        return driver.sd_ble_gattc_exchange_mtu_request(self.rpc_adapter,
                                                        conn_handle,
                                                        att_mtu_size)

    @NordicSemiErrorCheck
    @wrapt.synchronized(api_lock)
    def ble_gattc_hv_confirm(self, conn_handle, attr_handle):
        return driver.sd_ble_gattc_hv_confirm(self.rpc_adapter, conn_handle, attr_handle)

    """
    Driver handlers
    """

    def _status_handler(self, adapter, status_code, status_message):
        # print(status_message)
        pass

    def _log_message_handler(self, adapter, severity, log_message):
        if self._log_driver_comms:
            logger.info("LOG [{}]: {}".format(severity, log_message))

    """
    Event handling
    """

    def ble_evt_handler(self, adapter, ble_event):
        self._events.put(ble_event)

    def _event_handler(self):
        self._event_loop = True
        self._event_stopped.clear()
        while self._event_loop:
            try:
                ble_event = self._events.get(timeout=0.1)
            except queue.Empty:
                continue

            # logger.info('ble_event.header.evt_id %r', ble_event.header.evt_id)

            if len(self.observers) == 0:
                continue

            event = event_decode(ble_event)
            if event is None:
                logger.warning('unknown ble_event %r (discarded)', ble_event.header.evt_id)
                continue

            # logger.debug('ble_event.header.evt_id %r ----  %r', ble_event.header.evt_id, event)

            # Get a copy of the observers and event observers in case its modified during this execution
            with self._event_observer_lock:
                observers = self.observers[:]
                event_handlers = self._event_observers.copy()

            # Call all the observers
            for obs in observers:
                try:
                    obs.on_driver_event(self, event)
                except:
                    traceback.print_exc()

            # Call all the handlers for the event type provided
            for event_type, handlers in event_handlers.items():
                if issubclass(type(event), event_type):
                    for handler in handlers:
                        try:
                            handler(self, event)
                        except:
                            traceback.print_exc()

        self._event_stopped.set()



"""
This example exhibits some of the functionality of a peripheral BLE device,
such as reading, writing and notifying characteristics.

This peripheral can be used with one of the central examples running on a separate nordic device,
or can be run with the nRF Connect app to explore the contents of the service
"""
import atexit
import binascii
import struct
import threading
import time

from blatann import BleDevice
from blatann.examples import example_utils, constants
from blatann.gap import advertising, smp, IoCapabilities
from blatann.waitables import GenericWaitable

logger = example_utils.setup_logger(level="DEBUG")


def on_connect(peer, event_args):
    """
    Event callback for when a central device connects to us

    :param peer: The peer that connected to us
    :type peer: blatann.peer.Client
    :param event_args: None
    """
    if peer:
        logger.info("Connected to peer")
    else:
        logger.warning("Connection timed out")


def on_disconnect(peer, event_args):
    """
    Event callback for when the client disconnects from us (or when we disconnect from the client)

    :param peer: The peer that disconnected
    :type peer: blatann.peer.Client
    :param event_args: The event args
    :type event_args: blatann.event_args.DisconnectionEventArgs
    """
    logger.info("Disconnected from peer, reason: {}".format(event_args.reason))


def on_hex_conversion_characteristic_write(characteristic, event_args):
    """
    Event callback for when the client writes to the hex conversion characteristic.
    This takes the data written, converts it to the hex representation, and updates the characteristic
    with this new value. If the client is subscribed to the characteristic, the client will be notified.

    :param characteristic: The hex conversion characteristic
    :type characteristic: blatann.gatt.gatts.GattsCharacteristic
    :param event_args: the event arguments
    :type event_args: blatann.event_args.WriteEventArgs
    """
    new_value = binascii.hexlify(event_args.value)
    logger.info("Got characteristic write - characteristic: {}, data: 0x{}".format(characteristic.uuid, new_value))
    characteristic.set_value(new_value[:characteristic.max_length], notify_client=True)


def on_gatts_subscription_state_changed(characteristic, event_args):
    """
    Event callback for when a client subscribes or unsubscribes from a characteristic. This
    is the equivalent to when a client writes to a CCCD descriptor on a characteristic.

    :type characteristic: blatann.gatt.gatts.GattsCharacteristic
    :type event_args: blatann.event_args.SubscriptionStateChangeEventArgs
    """
    logger.info("Subscription state changed - characteristic: {}, state: {}".format(
        characteristic.uuid, event_args.subscription_state))


def on_time_char_read(characteristic, event_args):
    """
    Event callback for when the client reads our time characteristic. Gets the current time and updates the characteristic.
    This demonstrates "lazy evaluation" of characteristics--instead of having to constantly update this characteristic,
    it is only updated when read/observed by an outside actor.

    :param characteristic: the time characteristic
    :type characteristic: blatann.gatt.gatts.GattsCharacteristic
    :param event_args: None
    """
    t = time.time()
    ms = int((t * 1000) % 1000)
    msg = "Time: {}.{:03}".format(time.strftime("%H:%M:%S", time.localtime(t)), ms)
    characteristic.set_value(msg)


def on_client_pairing_complete(peer, event_args):
    """
    Event callback for when the pairing process completes with the client

    :param peer: the peer that completed pairing
    :type peer: blatann.peer.Client
    :param event_args: the event arguments
    :type event_args: blatann.event_args.PairingCompleteEventArgs
    """
    logger.info("Client Pairing complete, status: {}".format(event_args.status))


def on_passkey_display(peer, event_args):
    """
    Event callback that is called when a passkey is required to be displayed to a user
    for the pairing process.

    :param peer: The peer the passkey is for
    :type peer: blatann.peer.Client
    :param event_args: The event args
    :type event_args: blatann.event_args.PasskeyDisplayEventArgs
    """
    logger.info("Passkey display: {}, match: {}".format(event_args.passkey, event_args.match_request))
    if event_args.match_request:
        response = input("Passkey: {}, do both devices show same passkey? [y/n]\n".format(event_args.passkey))
        match = response.lower().startswith("y")
        event_args.match_confirm(match)


def on_passkey_entry(peer, passkey_event_args):
    """
    Callback for when the user is requested to enter a passkey to resume the pairing process.
    Requests the user to enter the passkey and resolves the event with the passkey entered

    :param peer: the peer the passkey is for
    :param passkey_event_args:
    :type passkey_event_args: blatann.event_args.PasskeyEntryEventArgs
    """
    passkey = input("Enter passkey: ")
    passkey_event_args.resolve(passkey)


class CountingCharacteristicThread(object):
    """
    Thread which updates the counting characteristic and notifies
    the client each time its updated.
    This also demonstrates the notification queuing functionality--if a notification/indication
    is already in progress, future notifications will be queued and sent out when the previous ones complete.
    """
    def __init__(self, characteristic):
        """
        :param characteristic: the counting characteristic
        :type characteristic: blatann.gatt.gatts.GattsCharacteristic
        """
        self.current_value = 0
        self._stop_event = threading.Event()
        self._stopped = threading.Event()
        self.characteristic = characteristic
        self.characteristic.on_notify_complete.register(self._on_notify_complete)
        self.thread = threading.Thread(target=self.run)
        atexit.register(self.join)
        self.thread.daemon = True
        self.thread.start()

    def join(self):
        """
        Used to stop and join the thread
        """
        self._stop_event.set()
        self._stopped.wait(3)

    def _on_notify_complete(self, characteristic, event_args):
        """
        Event callback that is triggered when the notification finishes sending

        :param characteristic: The characteristic the notification was on
        :type characteristic: blatann.gatt.gatts.GattsCharacteristic
        :param event_args: The event arguments
        :type event_args: blatann.event_args.NotificationCompleteEventArgs
        """
        logger.info("Notification Complete, id: {}, reason: {}".format(event_args.id, event_args.reason))

    def run(self):
        while not self._stop_event.is_set():
            try:
                if not self.characteristic.client_subscribed:  # Do nothing until a client is subscribed
                    time.sleep(1)
                    continue
                # Increment the value and pack it
                self.current_value += 1
                value = struct.pack("<I", self.current_value)

                # Send out a notification of this new value
                waitable = self.characteristic.notify(value)
                # Send a burst of 16, then wait for them all to send before trying to send more
                if self.current_value % 16 == 0:
                    waitable.wait()
                    time.sleep(1)  # Wait a second before sending out the next burst
            except Exception as e:
                logger.exception(e)

        self._stopped.set()


def main(serial_port):
    # Create and open the device
    ble_device = BleDevice(serial_port)
    ble_device.configure()
    ble_device.open()

    # Set up desired security parameters
    ble_device.client.security.set_security_params(passcode_pairing=False, bond=False, lesc_pairing=False,
                                                   io_capabilities=IoCapabilities.DISPLAY_ONLY, out_of_band=False)
    ble_device.client.security.on_pairing_complete.register(on_client_pairing_complete)
    ble_device.client.security.on_passkey_display_required.register(on_passkey_display)
    ble_device.client.security.on_passkey_required.register(on_passkey_entry)

    # Create and add the math service
    service = ble_device.database.add_service(constants.MATH_SERVICE_UUID)

    # Create and add the hex conversion characteristic to the service
    hex_conv_char = service.add_characteristic(constants.HEX_CONVERT_CHAR_UUID,
                                               constants.HEX_CONVERT_CHAR_PROPERTIES, "Test Data")
    # Register the callback for when a write occurs and subscription state changes
    hex_conv_char.on_write.register(on_hex_conversion_characteristic_write)
    hex_conv_char.on_subscription_change.register(on_gatts_subscription_state_changed)

    # Create and add the counting characteristic, initializing the data to [0, 0, 0, 0]
    counting_char = service.add_characteristic(constants.COUNTING_CHAR_UUID, constants.COUNTING_CHAR_PROPERTIES, [0]*4)
    counting_char.on_subscription_change.register(on_gatts_subscription_state_changed)

    # Create the thread for the counting characteristic
    counting_char_thread = CountingCharacteristicThread(counting_char)

    # Create and add the time service
    time_service = ble_device.database.add_service(constants.TIME_SERVICE_UUID)

    # Add the time characteristic and register the callback for when its read
    time_char = time_service.add_characteristic(constants.TIME_CHAR_UUID, constants.TIME_CHAR_PROPERTIES, "Time")
    time_char.on_read.register(on_time_char_read)

    # Initialize the advertising and scan response data
    adv_data = advertising.AdvertisingData(local_name=constants.PERIPHERAL_NAME, flags=0x06)
    scan_data = advertising.AdvertisingData(service_uuid128s=constants.TIME_SERVICE_UUID, has_more_uuid128_services=True)
    ble_device.advertiser.set_advertise_data(adv_data, scan_data)

    # Start advertising
    logger.info("Advertising")
    ble_device.client.on_connect.register(on_connect)
    ble_device.client.on_disconnect.register(on_disconnect)
    ble_device.advertiser.start(timeout_sec=0, auto_restart=True)

    # Create a waitable that will never fire, and wait for some time
    w = GenericWaitable()
    w.wait(60*30, exception_on_timeout=False)  # Keep device active for 30 mins

    # Cleanup
    counting_char_thread.join()
    logger.info("Done")
    ble_device.close()
    

if __name__ == '__main__':
    main("COM8")



from blatann.services.nordic_uart.service import NordicUartServer as _Server, NordicUartClient as _Client
from blatann.services.nordic_uart.constants import (
    NORDIC_UART_SERVICE_UUID,
    NORDIC_UART_RX_CHARACTERISTIC_UUID,
    NORDIC_UART_TX_CHARACTERISTIC_UUID,
    NORDIC_UART_FEATURE_CHARACTERISTIC_UUID
)


def add_nordic_uart_service(gatts_database, max_characteristic_size=None):
    """
    Adds a Nordic UART service to the database

    :param gatts_database: The database to add the service to
    :type gatts_database: blatann.gatt.gatts.GattsDatabase
    :param max_characteristic_size: The size of the characteristic which determines the read/write chunk size.
                                    This should be tuned to the MTU size of the connection
    :return: The Nordic Uart Service
    :rtype: _Server
    """
    return _Server.add_to_database(gatts_database, max_characteristic_size)


def find_nordic_uart_service(gattc_database):
    """
    Finds a Nordic UART service in the given GATT client database

    :param gattc_database: the GATT client database to search
    :type gattc_database: blatann.gatt.gattc.GattcDatabase
    :return: The UART service if found, None if not found
    :rtype: _Client
    """
    return _Client.find_in_database(gattc_database)



from enum import Enum
import logging
from blatann.nrf.nrf_dll_load import driver
import blatann.nrf.nrf_driver_types as util

logger = logging.getLogger(__name__)

BLE_CONN_HANDLE_INVALID = driver.BLE_CONN_HANDLE_INVALID

NoneType = type(None)


class BLEUUIDBase(object):
    BLE_UUID_TYPE_BLE = driver.BLE_UUID_TYPE_BLE

    def __init__(self, vs_uuid_base=None, uuid_type=None):
        assert isinstance(vs_uuid_base, (list, NoneType)), 'Invalid argument type'
        assert isinstance(uuid_type, (int, NoneType)), 'Invalid argument type'
        if vs_uuid_base is None:
            self.base = [0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x00,
                         0x80, 0x00, 0x00, 0x80, 0x5F, 0x9B, 0x34, 0xFB]
            self.def_base = True
        else:
            self.base = vs_uuid_base
            self.def_base = False

        if uuid_type is None and self.def_base:
            self.type = driver.BLE_UUID_TYPE_BLE
        else:
            self.type = uuid_type if uuid_type is not None else 0

    def __eq__(self, other):
        if not isinstance(other, BLEUUIDBase):
            return False
        if self.base != other.base:
            return False
        if self.type != other.type:
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    @classmethod
    def from_c(cls, uuid):
        if uuid.type == driver.BLE_UUID_TYPE_BLE:
            return cls(uuid_type=uuid.type)
        else:
            return cls([0] * 16, uuid_type=uuid.type)  # TODO: Hmmmm? [] or [None]*16? what?

    @classmethod
    def from_uuid128_array(cls, uuid128_array):
        msb_list = uuid128_array[::-1]
        return cls(msb_list)

    def to_c(self):
        lsb_list = self.base[::-1]
        self.__array = util.list_to_uint8_array(lsb_list)
        uuid = driver.ble_uuid128_t()
        uuid.uuid128 = self.__array.cast()
        return uuid


class BLEUUID(object):
    class Standard(Enum):
        unknown = 0x0000
        service_primary = 0x2800
        service_secondary = 0x2801
        characteristic = 0x2803
        cccd = 0x2902
        battery_level = 0x2A19
        heart_rate = 0x2A37

    def __init__(self, value, base=BLEUUIDBase()):
        assert isinstance(base, BLEUUIDBase), 'Invalid argument type'
        self.base = base
        if self.base.def_base:
            try:
                self.value = value if isinstance(value, BLEUUID.Standard) else BLEUUID.Standard(value)
            except ValueError:
                self.value = value
        else:
            self.value = value

    def get_value(self):
        if isinstance(self.value, BLEUUID.Standard):
            return self.value.value
        return self.value

    def as_array(self):
        base_and_value = self.base.base[:]
        base_and_value[2] = (self.get_value() >> 8) & 0xff
        base_and_value[3] = (self.get_value() >> 0) & 0xff
        return base_and_value

    def __str__(self):
        if isinstance(self.value, BLEUUID.Standard):
            return '0x{:04X} ({})'.format(self.value.value, self.value)
        elif self.base.type == driver.BLE_UUID_TYPE_BLE and self.base.def_base:
            return '0x{:04X}'.format(self.value)
        else:
            base_and_value = self.base.base[:]
            base_and_value[2] = (self.value >> 8) & 0xff
            base_and_value[3] = (self.value >> 0) & 0xff
            return '0x{}'.format(''.join(['{:02X}'.format(i) for i in base_and_value]))

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        if isinstance(other, BLEUUID.Standard):
            return self.get_value() == other.value
        if not isinstance(other, BLEUUID):
            return False
        if not self.base == other.base:
            return False
        if not self.value == other.value:
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(str(self))

    @classmethod
    def from_c(cls, uuid):
        return cls(value=uuid.uuid, base=BLEUUIDBase.from_c(uuid))  # TODO: Is this correct?

    @classmethod
    def from_uuid128(cls, uuid128):
        uuid = util.uint8_array_to_list(uuid128.uuid, 16)
        return cls.from_array(uuid)

    def to_c(self):
        assert self.base.type is not None, 'Vendor specific UUID not registered'
        uuid = driver.ble_uuid_t()
        if isinstance(self.value, BLEUUID.Standard):
            uuid.uuid = self.value.value
        else:
            uuid.uuid = self.value
        uuid.type = self.base.type
        return uuid

    @classmethod
    def from_array(cls, uuid_array_lt):
        base = list(reversed(uuid_array_lt))
        uuid = (base[2] << 8) + base[3]
        base[2] = 0
        base[3] = 0
        return cls(value=uuid, base=BLEUUIDBase(base, 0))



from typing import TypeVar, Generic, Callable
from threading import Lock

TSender = TypeVar("TSender")
TEvent = TypeVar("TEvent")


class Event(Generic[TSender, TEvent]):
    """
    Represents an event that can have handlers registered and deregistered.
    All handlers registered to an event should take in two parameters: the event sender and the event arguments.
    Those familiar with the C#/.NET event architecture, this should look very similar.
    """
    def __init__(self, name):
        self.name = name
        self._handler_lock = Lock()
        self._handlers = []

    def register(self, handler: Callable[[TSender, TEvent], None]):
        """
        Registers a handler to be called whenever the event is emitted.
        If the given handler is already registered, function does nothing

        :param handler: The handler to register
        """
        with self._handler_lock:
            if handler not in self._handlers:
                self._handlers.append(handler)

    def deregister(self, handler):
        """
        Deregisters a previously-registered handler so it no longer receives the event.
        If the given handler is not registered, function does nothing

        :param handler: The handler to deregister
        """
        with self._handler_lock:
            if handler in self._handlers:
                self._handlers.remove(handler)


class EventSource(Event):
    """
    Represents an Event object along with the controls to emit the events and notify handlers.
    This is done to "hide" the notify method from subscribers.
    """
    def __init__(self, name, logger=None):
        super(EventSource, self).__init__(name)
        self._logger = logger

    def clear_handlers(self):
        with self._handler_lock:
            self._handlers = []

    def notify(self, sender: TSender, event_args: TEvent = None):
        """
        Notifies all clients with the given arguments and keyword-arguments
        """
        with self._handler_lock:
            handlers = self._handlers[:]

        for h in handlers:
            try:
                h(sender, event_args)
            except Exception as e:
                if self._logger:
                    self._logger.exception(e)



from enum import IntEnum

from blatann.nrf.nrf_types import *
from blatann.nrf.nrf_dll_load import driver
import blatann.nrf.nrf_driver_types as util
from blatann.nrf.nrf_events.generic_events import BLEEvent


"""
Base Event Classes
"""


class GattEvt(BLEEvent):
    pass


class GattcEvt(GattEvt):
    pass


class GattsEvt(GattEvt):
    pass


"""
GATTC Events
"""


class GattcEvtReadResponse(GattcEvt):
    evt_id = driver.BLE_GATTC_EVT_READ_RSP

    def __init__(self, conn_handle, status, error_handle, attr_handle, offset, data):
        super(GattcEvtReadResponse, self).__init__(conn_handle)
        self.status = status
        self.error_handle = error_handle
        self.attr_handle = attr_handle
        self.offset = offset
        if status == BLEGattStatusCode.read_not_permitted:
            self.data = None
        elif isinstance(data, str):
            self.data = list(map(ord, data))
        else:
            self.data = data

    @classmethod
    def from_c(cls, event):
        read_rsp = event.evt.gattc_evt.params.read_rsp
        return cls(conn_handle=event.evt.gattc_evt.conn_handle,
                   status=BLEGattStatusCode(event.evt.gattc_evt.gatt_status),
                   error_handle=event.evt.gattc_evt.error_handle,
                   attr_handle=read_rsp.handle,
                   offset=read_rsp.offset,
                   data=util.uint8_array_to_list(read_rsp.data, read_rsp.len))

    def __repr__(self):
        data = None
        if self.data is not None:
            data = ''.join(map(chr, self.data))
        return "{}(conn_handle={!r}, status={!r}, error_handle={!r}, attr_handle={!r}, offset={!r}, data={!r})".format(
            self.__class__.__name__, self.conn_handle,
            self.status, self.error_handle, self.attr_handle, self.offset, data)


class GattcEvtHvx(GattcEvt):
    evt_id = driver.BLE_GATTC_EVT_HVX

    def __init__(self, conn_handle, status, error_handle, attr_handle, hvx_type, data):
        super(GattcEvtHvx, self).__init__(conn_handle)
        self.status = status
        self.error_handle = error_handle
        self.attr_handle = attr_handle
        self.hvx_type = hvx_type
        if isinstance(data, str):
            self.data = list(map(ord, data))
        else:
            self.data = data

    @classmethod
    def from_c(cls, event):
        hvx_evt = event.evt.gattc_evt.params.hvx
        return cls(conn_handle=event.evt.gattc_evt.conn_handle,
                   status=BLEGattStatusCode(event.evt.gattc_evt.gatt_status),
                   error_handle=event.evt.gattc_evt.error_handle,
                   attr_handle=hvx_evt.handle,
                   hvx_type=BLEGattHVXType(hvx_evt.type),
                   data=util.uint8_array_to_list(hvx_evt.data, hvx_evt.len))

    def __repr__(self):
        data = ''.join(map(chr, self.data))
        return "{}(conn_handle={!r}, status={!r}, error_handle={!r}, attr_handle={!r}, hvx_type={!r}, data={!r})".format(
            self.__class__.__name__, self.conn_handle,
            self.status, self.error_handle, self.attr_handle, self.hvx_type, data)


class GattcEvtWriteResponse(GattcEvt):
    evt_id = driver.BLE_GATTC_EVT_WRITE_RSP

    def __init__(self, conn_handle, status, error_handle, attr_handle, write_op, offset, data):
        super(GattcEvtWriteResponse, self).__init__(conn_handle)
        self.status = status
        self.error_handle = error_handle
        self.attr_handle = attr_handle
        self.write_op = write_op
        self.offset = offset
        if isinstance(data, str):
            self.data = list(map(ord, data))
        else:
            self.data = data

    @classmethod
    def from_c(cls, event):
        write_rsp_evt = event.evt.gattc_evt.params.write_rsp
        return cls(conn_handle=event.evt.gattc_evt.conn_handle,
                   status=BLEGattStatusCode(event.evt.gattc_evt.gatt_status),
                   error_handle=event.evt.gattc_evt.error_handle,
                   attr_handle=write_rsp_evt.handle,
                   write_op=BLEGattWriteOperation(write_rsp_evt.write_op),
                   offset=write_rsp_evt.offset,
                   data=util.uint8_array_to_list(write_rsp_evt.data, write_rsp_evt.len))

    def __repr__(self):
        data = ''.join(map(chr, self.data))
        return "{}(conn_handle={!r}, status={!r}, error_handle={!r}, attr_handle={!r}, write_op={!r}, offset={!r}, data={!r})".format(
            self.__class__.__name__, self.conn_handle,
            self.status, self.error_handle, self.attr_handle, self.write_op, self.offset, data)


class GattcEvtPrimaryServiceDiscoveryResponse(GattcEvt):
    evt_id = driver.BLE_GATTC_EVT_PRIM_SRVC_DISC_RSP

    def __init__(self, conn_handle, status, services):
        super(GattcEvtPrimaryServiceDiscoveryResponse, self).__init__(conn_handle)
        self.status = status
        self.services = services

    @classmethod
    def from_c(cls, event):
        prim_srvc_disc_rsp_evt = event.evt.gattc_evt.params.prim_srvc_disc_rsp

        services = list()
        for s in util.service_array_to_list(prim_srvc_disc_rsp_evt.services, prim_srvc_disc_rsp_evt.count):
            services.append(BLEGattService.from_c(s))

        return cls(conn_handle=event.evt.gattc_evt.conn_handle,
                   status=BLEGattStatusCode(event.evt.gattc_evt.gatt_status),
                   services=services)

    def __repr__(self):
        return "{}(conn_handle={!r}, status={!r}, services={!r})".format(self.__class__.__name__, self.conn_handle,
                                                                         self.status, self.services)


class GattcEvtCharacteristicDiscoveryResponse(GattcEvt):
    evt_id = driver.BLE_GATTC_EVT_CHAR_DISC_RSP

    def __init__(self, conn_handle, status, characteristics):
        super(GattcEvtCharacteristicDiscoveryResponse, self).__init__(conn_handle)
        self.status = status
        self.characteristics = characteristics

    @classmethod
    def from_c(cls, event):
        char_disc_rsp_evt = event.evt.gattc_evt.params.char_disc_rsp

        characteristics = list()
        for ch in util.ble_gattc_char_array_to_list(char_disc_rsp_evt.chars, char_disc_rsp_evt.count):
            characteristics.append(BLEGattCharacteristic.from_c(ch))

        return cls(conn_handle=event.evt.gattc_evt.conn_handle,
                   status=BLEGattStatusCode(event.evt.gattc_evt.gatt_status),
                   characteristics=characteristics)

    def __repr__(self):
        return "{}(conn_handle={!r}, status={!r}, characteristics={!r})".format(self.__class__.__name__,
                                                                                self.conn_handle,
                                                                                self.status, self.characteristics)


class GattcEvtDescriptorDiscoveryResponse(GattcEvt):
    evt_id = driver.BLE_GATTC_EVT_DESC_DISC_RSP

    def __init__(self, conn_handle, status, descriptions):
        super(GattcEvtDescriptorDiscoveryResponse, self).__init__(conn_handle)
        self.status = status
        self.descriptions = descriptions

    @classmethod
    def from_c(cls, event):
        desc_disc_rsp_evt = event.evt.gattc_evt.params.desc_disc_rsp

        descriptions = list()
        for d in util.desc_array_to_list(desc_disc_rsp_evt.descs, desc_disc_rsp_evt.count):
            descriptions.append(BLEGattcDescriptor.from_c(d))

        return cls(conn_handle=event.evt.gattc_evt.conn_handle,
                   status=BLEGattStatusCode(event.evt.gattc_evt.gatt_status),
                   descriptions=descriptions)

    def __repr__(self):
        return "{}(conn_handle={!r}, status={!r}, descriptions={!r})".format(self.__class__.__name__, self.conn_handle,
                                                                             self.status, self.descriptions)


class GattcEvtAttrInfoDiscoveryResponse(GattcEvt):
    evt_id = driver.BLE_GATTC_EVT_ATTR_INFO_DISC_RSP

    def __init__(self, conn_handle, status, attr_info16=None, attr_info128=None):
        super(GattcEvtAttrInfoDiscoveryResponse, self).__init__(conn_handle)
        self.status = status
        self.attr_info16 = attr_info16
        self.attr_info128 = attr_info128

    @classmethod
    def from_c(cls, event):
        status = event.evt.gattc_evt.gatt_status
        attr_info_rsp = event.evt.gattc_evt.params.attr_info_disc_rsp
        if attr_info_rsp.format == driver.BLE_GATTC_ATTR_INFO_FORMAT_16BIT:
            attr16_array = util.attr_info16_array_to_list(attr_info_rsp.info.attr_info16, attr_info_rsp.count)
            attr16 = [BLEGattcAttrInfo16.from_c(a) for a in attr16_array]
            attr128 = None
        else:
            attr16 = None
            attr128_array = util.attr_info128_array_to_list(attr_info_rsp.info.attr_info128, attr_info_rsp.count)
            attr128 = [BLEGattcAttrInfo128.from_c(a) for a in attr128_array]

        return cls(event.evt.gattc_evt.conn_handle, status, attr16, attr128)

    def __repr__(self):
        return "{}(conn_handle={!r}, status: {!r} uuid_info:{})".format(self.__class__.__name__,
                                                                        self.conn_handle, self.status,
                                                                        self.attr_info16 or self.attr_info128)


class GattcEvtMtuExchangeResponse(GattcEvt):
    evt_id = driver.BLE_GATTC_EVT_EXCHANGE_MTU_RSP

    def __init__(self, conn_handle, server_mtu):
        super(GattcEvtMtuExchangeResponse, self).__init__(conn_handle)
        self.server_mtu = server_mtu

    @classmethod
    def from_c(cls, event):
        server_mtu_size = event.evt.gattc_evt.params.exchange_mtu_rsp.server_rx_mtu
        return cls(event.evt.gattc_evt.conn_handle, server_mtu_size)

    def __repr__(self):
        return "{}(conn_handle={!r}, server_mtu={})".format(self.__class__.__name__, self.conn_handle, self.server_mtu)


class GattcEvtTimeout(GattcEvt):
    evt_id = driver.BLE_GATTC_EVT_TIMEOUT

    def __init__(self, conn_handle, source):
        super(GattcEvtTimeout, self).__init__(conn_handle)
        self.source = source

    @classmethod
    def from_c(cls, event):
        source = event.evt.gattc_evt.params.timeout.src
        return cls(event.evt.gattc_evt.conn_handle, source)

    def __repr__(self):
        return self._repr_format(source=self.source)

"""
GATTS Events
"""


# TODO: SYS_ATTR_MISSING, SC_CONFIRM, TIMEOUT


class GattsEvtWrite(GattsEvt):
    evt_id = driver.BLE_GATTS_EVT_WRITE

    def __init__(self, conn_handle, attr_handle, uuid, write_operand, auth_required, offset, data):
        super(GattsEvtWrite, self).__init__(conn_handle)
        self.attribute_handle = attr_handle
        self.uuid = uuid
        self.write_op = write_operand
        self.auth_required = auth_required
        self.offset = offset
        self.data = data

    @classmethod
    def from_c(cls, event):
        conn_handle = event.evt.gatts_evt.conn_handle
        write_event = event.evt.gatts_evt.params.write
        return cls.from_auth_request(conn_handle, write_event)

    @classmethod
    def from_auth_request(cls, conn_handle, write_event):
        attr_handle = write_event.handle
        uuid = BLEUUID.from_c(write_event.uuid)
        write_operand = BLEGattsWriteOperation(write_event.op)
        auth_required = bool(write_event.auth_required)
        offset = write_event.offset
        data = util.uint8_array_to_list(write_event.data, write_event.len)

        return cls(conn_handle, attr_handle, uuid, write_operand, auth_required, offset, data)

    def __repr__(self):
        return "{}(conn_handle={!r}, attr_handle={!r}, uuid={!r}, write_op={!r}, auth_required={!r}, offset={!r}, " \
               "data={!r})".format(self.__class__.__name__, self.conn_handle, self.attribute_handle, self.uuid,
                                   self.write_op, self.auth_required, self.offset, self.data)


class GattsEvtRead(GattsEvt):  # Not a _true_ event, but is used for the Read/Write event class
    def __init__(self, conn_handle, attr_handle, uuid, offset):
        super(GattsEvtRead, self).__init__(conn_handle)
        self.attribute_handle = attr_handle
        self.uuid = uuid
        self.offset = offset

    @classmethod
    def from_auth_request(cls, conn_handle, read_event):
        attr_handle = read_event.handle
        uuid = BLEUUID.from_c(read_event.uuid)
        offset = read_event.offset

        return cls(conn_handle, attr_handle, uuid, offset)

    def __repr__(self):
        return "{}(conn_handle={!r}, attr_handle={!r}, uuid={!r}, offset={!r}, )".format(self.__class__.__name__,
                                                                                         self.conn_handle,
                                                                                         self.attribute_handle,
                                                                                         self.uuid, self.offset)


class GattsEvtReadWriteAuthorizeRequest(GattsEvt):
    evt_id = driver.BLE_GATTS_EVT_RW_AUTHORIZE_REQUEST

    def __init__(self, conn_handle, read=None, write=None):
        super(GattsEvtReadWriteAuthorizeRequest, self).__init__(conn_handle)
        self.read = read
        self.write = write

    @classmethod
    def from_c(cls, event):
        auth_event = event.evt.gatts_evt.params.authorize_request
        conn_handle = event.evt.gatts_evt.conn_handle
        read = None
        write = None
        if auth_event.type == driver.BLE_GATTS_AUTHORIZE_TYPE_READ:
            read = GattsEvtRead.from_auth_request(conn_handle, auth_event.request.read)
        elif auth_event.type == driver.BLE_GATTS_AUTHORIZE_TYPE_WRITE:
            write = GattsEvtWrite.from_auth_request(conn_handle, auth_event.request.write)
        else:
            raise NordicSemiException("Unknown authorize request type: {}".format(auth_event.type))
        return cls(conn_handle, read, write)

    def __repr__(self):
        if self.read is not None:
            txt = "read"
            data = self.read
        else:
            txt = "write"
            data = self.write

        return "{}(conn_handle={!r}, {}={!r})".format(self.__class__.__name__, self.conn_handle, txt, data)


class GattsEvtHandleValueConfirm(GattsEvt):
    evt_id = driver.BLE_GATTS_EVT_HVC

    def __init__(self, conn_handle, attr_handle):
        super(GattsEvtHandleValueConfirm, self).__init__(conn_handle)
        self.attribute_handle = attr_handle

    @classmethod
    def from_c(cls, event):
        conn_handle = event.evt.gatts_evt.conn_handle
        return cls(conn_handle, event.evt.gatts_evt.params.hvc.handle)

    def __repr__(self):
        return "{}(conn_handle={!r}, attr_handle={!r})".format(self.__class__.__name__, self.conn_handle, self.attribute_handle)


class GattsEvtNotificationTxComplete(GattsEvt):
    evt_id = driver.BLE_GATTS_EVT_HVN_TX_COMPLETE

    def __init__(self, conn_handle, tx_count):
        super(GattsEvtNotificationTxComplete, self).__init__(conn_handle)
        self.tx_count = tx_count

    @classmethod
    def from_c(cls, event):
        conn_handle = event.evt.gatts_evt.conn_handle
        return cls(conn_handle, event.evt.gatts_evt.params.hvn_tx_complete.count)

    def __repr__(self):
        return self._repr_format(tx_count=self.tx_count)


class GattsEvtExchangeMtuRequest(GattsEvt):
    evt_id = driver.BLE_GATTS_EVT_EXCHANGE_MTU_REQUEST

    def __init__(self, conn_handle, client_mtu):
        super(GattsEvtExchangeMtuRequest, self).__init__(conn_handle)
        self.client_mtu = client_mtu

    @classmethod
    def from_c(cls, event):
        conn_handle = event.evt.gatts_evt.conn_handle
        return cls(conn_handle, event.evt.gatts_evt.params.exchange_mtu_request.client_rx_mtu)

    def __repr__(self):
        return "{}(conn_handle={!r}, client_mtu={!r})".format(self.__class__.__name__, self.conn_handle, self.client_mtu)


class GattsEvtTimeout(GattsEvt):
    evt_id = driver.BLE_GATTS_EVT_TIMEOUT

    def __init__(self, conn_handle, source):
        super(GattsEvtTimeout, self).__init__(conn_handle)
        self.source = source

    @classmethod
    def from_c(cls, event):
        source = event.evt.gatts_evt.params.timeout.src
        return cls(event.evt.gatts_evt.conn_handle, source)

    def __repr__(self):
        return self._repr_format(source=self.source)


import threading
import queue
import logging


logger = logging.getLogger(__name__)


class QueuedTasksManagerBase(object):
    """
    Handles queuing of tasks that can only be done one at a time
    """
    def __init__(self):
        self._queue = queue.Queue()
        self._in_process = threading.Event()
        self._lock = threading.RLock()

    def _add_task(self, task):
        with self._lock:
            if self._in_process.is_set():
                self._queue.put(task)
            else:
                try:
                    task_complete = self._handle_task(task)
                    if not task_complete:
                        self._in_process.set()
                except Exception as e:
                    self._handle_task_failure(task, e)

    def _task_completed(self, task):
        with self._lock:
            task = self._get_next()
            while task:
                try:
                    task_complete = self._handle_task(task)
                    if not task_complete:
                        break
                except Exception as e:
                    logger.exception(e)
                    self._handle_task_failure(task, e)
                task = self._get_next()

            if not task:
                self._in_process.clear()

    def _get_next(self):
        try:
            return self._queue.get_nowait()
        except queue.Empty:
            return None

    def _clear_all(self, reason):
        with self._lock:
            task = self._get_next()
            while task:
                self._handle_task_cleared(task, reason)
                task = self._get_next()
            self._in_process.clear()

    def clear_all(self):
        raise NotImplementedError()

    def _handle_task(self, task):
        raise NotImplementedError()

    def _handle_task_failure(self, task, e):
        raise NotImplementedError()

    def _handle_task_cleared(self, task, reason):
        raise NotImplementedError()




from blatann.uuid import Uuid16

# Uuids
DIS_SERVICE_UUID = Uuid16(0x180A)

SYSTEM_ID_UUID = Uuid16(0x2A23)
MODEL_NUMBER_UUID = Uuid16(0x2A24)
SERIAL_NUMBER_UUID = Uuid16(0x2A25)
FIRMWARE_REV_UUID = Uuid16(0x2A26)
HARDWARE_REV_UUID = Uuid16(0x2A27)
SOFTWARE_REV_UUID = Uuid16(0x2A28)
MANUFACTURER_NAME_UUID = Uuid16(0x2A29)
REGULATORY_CERT_UUID = Uuid16(0x2A2A)
PNP_ID_UUID = Uuid16(0x2A50)


class _Characteristic(object):
    def __init__(self, name, uuid):
        self.uuid = uuid
        self.name = name

    def __str__(self):
        return "{} ({})".format(self.name, self.uuid)


SystemIdCharacteristic = _Characteristic("System Id", SYSTEM_ID_UUID)
ModelNumberCharacteristic = _Characteristic("Model Number", MODEL_NUMBER_UUID)
SerialNumberCharacteristic = _Characteristic("Serial Number", SERIAL_NUMBER_UUID)
FirmwareRevisionCharacteristic = _Characteristic("Firmware Revision", FIRMWARE_REV_UUID)
HardwareRevisionCharacteristic = _Characteristic("Hardware Revision", HARDWARE_REV_UUID)
SoftwareRevisionCharacteristic = _Characteristic("Software Revision", SOFTWARE_REV_UUID)
ManufacturerNameCharacteristic = _Characteristic("Manufacturer Name", MANUFACTURER_NAME_UUID)
RegulatoryCertificateCharacteristic = _Characteristic("Regulatory Certificate", REGULATORY_CERT_UUID)
PnpIdCharacteristic = _Characteristic("PNP Id", PNP_ID_UUID)


CHARACTERISTICS = [
    SystemIdCharacteristic,
    ModelNumberCharacteristic,
    SerialNumberCharacteristic,
    FirmwareRevisionCharacteristic,
    HardwareRevisionCharacteristic,
    SoftwareRevisionCharacteristic,
    ManufacturerNameCharacteristic,
    RegulatoryCertificateCharacteristic,
    PnpIdCharacteristic
]




from queue import Queue
from typing import Iterable
from blatann.waitables.waitable import Waitable
from blatann.nrf.nrf_events import GapEvtTimeout, BLEGapTimeoutSrc, GapEvtAdvReport
from blatann.gap.advertise_data import ScanReport, ScanReportCollection


class ScanFinishedWaitable(Waitable):
    def __init__(self, ble_device):
        super(ScanFinishedWaitable, self).__init__()
        self.scanner = ble_device.scanner
        ble_device.ble_driver.event_subscribe(self._on_timeout_event, GapEvtTimeout)
        self.scanner.on_scan_received.register(self._on_scan_report)
        self.ble_driver = ble_device.ble_driver
        self._scan_report_queue = Queue()

    @property
    def scan_reports(self) -> Iterable[ScanReport]:
        scan_report = self._scan_report_queue.get()
        while scan_report:
            yield scan_report
            scan_report = self._scan_report_queue.get()

    def _event_occurred(self, ble_driver):
        ble_driver.event_unsubscribe(self._on_timeout_event, GapEvtTimeout)
        self.scanner.on_scan_received.deregister(self._on_scan_report)
        self._notify(self.scanner.scan_report)

    def _on_timeout(self):
        self.ble_driver.event_unsubscribe(self._on_timeout_event, GapEvtTimeout)
        self.scanner.on_scan_received.deregister(self._on_scan_report)

    def _on_scan_report(self, device, scan_report):
        self._scan_report_queue.put(scan_report)

    def _on_timeout_event(self, ble_driver, event):
        """
        :type event: GapEvtTimeout
        """
        if event.src == BLEGapTimeoutSrc.scan:
            self._event_occurred(ble_driver)
            self._scan_report_queue.put(None)

    def wait(self, timeout=None, exception_on_timeout=True) -> ScanReportCollection:
        return super(ScanFinishedWaitable, self).wait(timeout, exception_on_timeout)



import logging
from blatann.event_type import EventSource, Event
from blatann.nrf import nrf_types, nrf_events
from blatann.waitables.event_waitable import EventWaitable
from blatann.exceptions import InvalidStateException
from blatann.event_args import EventArgs

logger = logging.getLogger(__name__)


class GattcReadCompleteEventArgs(EventArgs):
    def __init__(self, handle, status, data):
        self.handle = handle
        self.status = status
        self.data = data


class GattcReader(object):
    """
    Class which implements the state machine for completely reading a peripheral's attribute
    """
    _READ_OVERHEAD = 1  # Number of bytes per MTU that are overhead for the read operation

    def __init__(self, ble_device, peer):
        """
        :type ble_device: blatann.device.BleDevice
        :type peer: blatann.peer.Peer
        """
        self.ble_device = ble_device
        self.peer = peer
        self._on_read_complete_event = EventSource("On Read Complete", logger)
        self._busy = False
        self._data = bytearray()
        self._handle = 0x0000
        self._offset = 0
        self.peer.driver_event_subscribe(self._on_read_response, nrf_events.GattcEvtReadResponse)

    @property
    def on_read_complete(self):
        """
        Event that is emitted when a read completes on an attribute handle.

        Handler args: (int attribute_handle, gatt.GattStatusCode, bytes data_read)

        :return: an Event which can have handlers registered to and deregistered from
        :rtype: Event
        """
        return self._on_read_complete_event

    def read(self, handle):
        """
        Reads the attribute value from the handle provided. Can only read from a single attribute at a time. If a
        read is in progress, raises an InvalidStateException

        :param handle: the attribute handle to read
        :return: A waitable that will fire when the read finishes.
                 See on_read_complete for the values returned from the waitable
        :rtype: EventWaitable
        """
        if self._busy:
            raise InvalidStateException("Gattc Reader is busy")
        self._handle = handle
        self._offset = 0
        self._data = bytearray()
        logger.debug("Starting read from handle {}".format(handle))
        self._read_next_chunk()
        self._busy = True
        return EventWaitable(self.on_read_complete)

    def _read_next_chunk(self):
        self.ble_device.ble_driver.ble_gattc_read(self.peer.conn_handle, self._handle, self._offset)

    def _on_read_response(self, driver, event):
        """
        Handler for GattcEvtReadResponse

        :type event: nrf_events.GattcEvtReadResponse
        """
        if event.conn_handle != self.peer.conn_handle or event.attr_handle != self._handle:
            return
        if event.status != nrf_events.BLEGattStatusCode.success:
            self._complete(event.status)
            return

        bytes_read = len(event.data)
        self._data += bytearray(event.data)
        self._offset += bytes_read

        if bytes_read == (self.peer.mtu_size - self._READ_OVERHEAD):
            self._read_next_chunk()
        else:
            self._complete()

    def _complete(self, status=nrf_events.BLEGattStatusCode.success):
        self._busy = False
        event_args = GattcReadCompleteEventArgs(self._handle, status, bytes(self._data))
        self._on_read_complete_event.notify(self, event_args)



import queue
from blatann.exceptions import TimeoutError


class Waitable(object):
    def __init__(self):
        self._queue = queue.Queue()
        self._callback = None

    def wait(self, timeout=None, exception_on_timeout=True):
        try:
            results = self._queue.get(timeout=timeout)
            if len(results) == 1:
                return results[0]
            return results
        except queue.Empty:
            self._on_timeout()
            if exception_on_timeout:
                raise TimeoutError("Timed out waiting for event to occur. "
                                   "Waitable type: {}".format(self.__class__.__name__))
        # TODO: This will fail if the waitable implementation normally returns more than one value and
        #       the caller tries to unpack
        return None

    def then(self, func_to_execute):
        self._callback = func_to_execute
        return self

    def _on_timeout(self):
        pass

    def _notify(self, *results):
        self._queue.put(results)
        if self._callback:
            self._callback(*results)


class GenericWaitable(Waitable):
    def notify(self, *results):
        self._notify(*results)



import binascii
import logging
from blatann.services import ble_data_types
from blatann.services.device_info.constants import *
from blatann.services.device_info.data_types import PnpId, SystemId
from blatann.event_type import EventSource
from blatann.event_args import DecodedReadCompleteEventArgs
from blatann.waitables import EventWaitable
from blatann.gatt import GattStatusCode
from blatann.gatt.gattc import GattcService
from blatann.gatt.gatts import GattsService, GattsCharacteristicProperties


logger = logging.getLogger(__name__)


class _DisCharacteristic(object):
    def __init__(self, service, uuid, data_class):
        self.service = service
        self.uuid = uuid
        self._char = None
        self.data_class = data_class

    @property
    def is_defined(self):
        return self._char is not None

    @property
    def value(self):
        if self._char:
            return self._char.value
        return None


class _DisServerCharacteristic(_DisCharacteristic):
    def set_value(self, value, max_len=None):
        if not self._char:
            props = GattsCharacteristicProperties(read=True, max_length=max_len or len(value))
            self._char = self.service.add_characteristic(self.uuid, props, value)
        else:
            self._char.set_value(value)


class _DisClientCharacteristic(_DisCharacteristic):
    def __init__(self, service, uuid, data_class):
        super(_DisClientCharacteristic, self).__init__(service, uuid, data_class)
        self._char = service.find_characteristic(uuid)
        self._on_read_complete_event = EventSource("Char {} Read Complete".format(self.uuid))

    def _read_complete(self, characteristic, event_args):
        """
        :param characteristic:
        :type event_args: blatann.event_args.ReadCompleteEventArgs
        """
        decoded_value = None
        if event_args.status == GattStatusCode.success:
            try:
                stream = ble_data_types.BleDataStream(event_args.value)
                decoded_value = self.data_class.decode(stream)
            except Exception as e:  # TODO not so generic
                logger.error("Service {}, Characteristic {} failed to decode value on read. "
                             "Stream: [{}]".format(self.service.uuid, self.uuid, binascii.hexlify(event_args.value)))
                logger.exception(e)

        decoded_event_args = DecodedReadCompleteEventArgs.from_read_complete_event_args(event_args, decoded_value)
        self._on_read_complete_event.notify(characteristic, decoded_event_args)

    def read(self):
        if not self.is_defined:
            raise AttributeError("Characteristic {} is not present in the Device Info Service".format(self.uuid))
        self._char.read().then(self._read_complete)
        return EventWaitable(self._on_read_complete_event)


class _DeviceInfoService(object):
    def __init__(self, service):
        if isinstance(service, GattsService):
            char_cls = _DisServerCharacteristic
        elif isinstance(service, GattcService):
            char_cls = _DisClientCharacteristic
        else:
            raise ValueError("Service must be a Gatt Server or Client")

        self._service = service
        self._system_id_char = char_cls(service, SYSTEM_ID_UUID, SystemId)
        self._model_number_char = char_cls(service, MODEL_NUMBER_UUID, ble_data_types.String)
        self._serial_no_char = char_cls(service, SERIAL_NUMBER_UUID, ble_data_types.String)
        self._firmware_rev_char = char_cls(service, FIRMWARE_REV_UUID, ble_data_types.String)
        self._hardware_rev_char = char_cls(service, HARDWARE_REV_UUID, ble_data_types.String)
        self._software_rev_char = char_cls(service, SOFTWARE_REV_UUID, ble_data_types.String)
        self._mfg_name_char = char_cls(service, MANUFACTURER_NAME_UUID, ble_data_types.String)
        self._regulatory_cert_char = char_cls(service, REGULATORY_CERT_UUID, ble_data_types.String)
        self._pnp_id_char = char_cls(service, PNP_ID_UUID, PnpId)

        self._characteristics = {
            SystemIdCharacteristic: self._system_id_char,
            ModelNumberCharacteristic: self._model_number_char,
            SerialNumberCharacteristic: self._serial_no_char,
            FirmwareRevisionCharacteristic: self._firmware_rev_char,
            HardwareRevisionCharacteristic: self._hardware_rev_char,
            SoftwareRevisionCharacteristic: self._software_rev_char,
            ManufacturerNameCharacteristic: self._mfg_name_char,
            RegulatoryCertificateCharacteristic: self._regulatory_cert_char,
            PnpIdCharacteristic: self._pnp_id_char
        }

    def has(self, characteristic):
        char = self._characteristics.get(characteristic, None)
        if not char:
            return False
        return char.is_defined

    @property
    def has_system_id(self):
        return self.has(SystemIdCharacteristic)

    @property
    def has_model_number(self):
        return self.has(ModelNumberCharacteristic)

    @property
    def has_serial_number(self):
        return self.has(SerialNumberCharacteristic)

    @property
    def has_firmware_revision(self):
        return self.has(FirmwareRevisionCharacteristic)

    @property
    def has_hardware_revision(self):
        return self.has(HardwareRevisionCharacteristic)

    @property
    def has_software_revision(self):
        return self.has(SoftwareRevisionCharacteristic)

    @property
    def has_manufacturer_name(self):
        return self.has(ManufacturerNameCharacteristic)

    @property
    def has_regulatory_certificate(self):
        return self.has(RegulatoryCertificateCharacteristic)

    @property
    def has_pnp_id(self):
        return self.has(PnpIdCharacteristic)


class DisClient(_DeviceInfoService):
    def __init__(self, gattc_service):
        """
        :type gattc_service: blatann.gatt.gattc.GattcService
        """
        super(DisClient, self).__init__(gattc_service)

    def get(self, characteristic):
        return self._characteristics[characteristic].read()

    def get_system_id(self):
        return self.get(SystemIdCharacteristic)

    def get_model_number(self):
        return self.get(ModelNumberCharacteristic)

    def get_serial_number(self):
        return self.get(SerialNumberCharacteristic)

    def get_firmware_revision(self):
        return self.get(FirmwareRevisionCharacteristic)

    def get_hardware_revision(self):
        return self.get(HardwareRevisionCharacteristic)

    def get_software_revision(self):
        return self.get(SoftwareRevisionCharacteristic)

    def get_manufacturer_name(self):
        return self.get(ManufacturerNameCharacteristic)

    def get_regulatory_certifications(self):
        return self.get(RegulatoryCertificateCharacteristic)

    def get_pnp_id(self):
        return self.get(PnpIdCharacteristic)

    @classmethod
    def find_in_database(cls, gattc_database):
        """
        :type gattc_database: blatann.gatt.gattc.GattcDatabase
        :rtype: DisClient
        """
        service = gattc_database.find_service(DIS_SERVICE_UUID)
        if service:
            return DisClient(service)


class DisServer(_DeviceInfoService):
    def set(self, characteristic, value, max_len=None):
        self._characteristics[characteristic].set_value(value, max_len)

    def set_system_id(self, system_id):
        """
        :type system_id: SystemId
        """
        if not isinstance(system_id, SystemId):
            raise ValueError("Value must be of type {}".format(SystemId.__name__))
        value = system_id.encode().value
        self.set(SystemIdCharacteristic, value)

    def set_model_number(self, model_number, max_len=None):
        self.set(ModelNumberCharacteristic, model_number, max_len)

    def set_serial_number(self, serial_number, max_len=None):
        self.set(SerialNumberCharacteristic, serial_number, max_len)

    def set_firmware_revision(self, firmware_revision, max_len=None):
        self.set(FirmwareRevisionCharacteristic, firmware_revision, max_len)

    def set_hardware_revision(self, hardware_revision, max_len=None):
        self.set(HardwareRevisionCharacteristic, hardware_revision, max_len)

    def set_software_revision(self, software_revision, max_len=None):
        self.set(SoftwareRevisionCharacteristic, software_revision, max_len)

    def set_manufacturer_name(self, manufacturer_name, max_len=None):
        self.set(ManufacturerNameCharacteristic, manufacturer_name, max_len)

    def set_regulatory_certifications(self, certs):
        raise NotImplementedError()

    def set_pnp_id(self, pnp_id):
        """
        :type pnp_id: PnpId
        """
        assert isinstance(pnp_id, PnpId)
        value = pnp_id.encode().value
        self.set(PnpIdCharacteristic, value)

    @classmethod
    def add_to_database(cls, gatts_database):
        """
        :type gatts_database: blatann.gatt.gatts.GattsDatabase
        """
        service = gatts_database.add_service(DIS_SERVICE_UUID)
        return DisServer(service)



from enum import Enum
import logging
import binascii
from blatann.nrf.nrf_dll_load import driver
import blatann.nrf.nrf_driver_types as util
from blatann.nrf.nrf_types.enums import *
from blatann.nrf.nrf_types.gap import BLEGapAddr


logger = logging.getLogger(__name__)


class BLEGapSecMode(object):
    def __init__(self, sec_mode, level):
        self.sm = sec_mode
        self.level = level

    def to_c(self):
        params = driver.ble_gap_conn_sec_mode_t()
        params.sm = self.sm
        params.lv = self.level
        return params

    @classmethod
    def from_c(cls, params):
        return cls(params.sm, params.lv)


class BLEGapSecModeType(object):
    NO_ACCESS = BLEGapSecMode(0, 0)
    OPEN = BLEGapSecMode(1, 1)
    ENCRYPTION = BLEGapSecMode(1, 2)
    MITM = BLEGapSecMode(1, 3)
    LESC_MITM = BLEGapSecMode(1, 4)
    SIGN_OR_ENCRYPT = BLEGapSecMode(2, 1)
    SIGN_OR_ENCRYPT_MITM = BLEGapSecMode(2, 2)


class BLEGapSecLevels(object):
    def __init__(self, lv1, lv2, lv3, lv4):
        self.lv1 = lv1
        self.lv2 = lv2
        self.lv3 = lv3
        self.lv4 = lv4

    @classmethod
    def from_c(cls, sec_level):
        return cls(lv1=sec_level.lv1,
                   lv2=sec_level.lv2,
                   lv3=sec_level.lv3,
                   lv4=sec_level.lv4)

    def to_c(self):
        sec_level = driver.ble_gap_sec_levels_t()
        sec_level.lv1 = self.lv1
        sec_level.lv2 = self.lv2
        sec_level.lv3 = self.lv3
        sec_level.lv4 = self.lv4
        return sec_level

    def __repr__(self):
        return "{}(lv1={!r}, lv2={!r}, lv3={!r}, lv4={!r})".format(self.__class__.__name__,
                                                                   self.lv1, self.lv2, self.lv3, self.lv4)


class BLEGapSecKeyDist(object):
    def __init__(self, enc_key=False, id_key=False, sign_key=False, link_key=False):
        self.enc_key = enc_key
        self.id_key = id_key
        self.sign_key = sign_key
        self.link_key = link_key

    @classmethod
    def from_c(cls, kdist):
        return cls(enc_key=kdist.enc,
                   id_key=kdist.id,
                   sign_key=kdist.sign,
                   link_key=kdist.link)

    def to_c(self):
        kdist = driver.ble_gap_sec_kdist_t()
        kdist.enc = self.enc_key
        kdist.id = self.id_key
        kdist.sign = self.sign_key
        kdist.link = self.link_key
        return kdist

    def __repr__(self):
        return "{}(enc_key={!r}, id_key={!r}, sign_key={!r}, link_key={!r})".format(
            self.__class__.__name__, self.enc_key, self.id_key, self.sign_key, self.link_key)


class BLEGapSecParams(object):
    def __init__(self, bond, mitm, le_sec_pairing, keypress_noti, io_caps, oob, min_key_size, max_key_size, kdist_own,
                 kdist_peer):
        self.bond = bond
        self.mitm = mitm
        self.le_sec_pairing = le_sec_pairing
        self.keypress_noti = keypress_noti
        self.io_caps = io_caps
        self.oob = oob
        self.min_key_size = min_key_size
        self.max_key_size = max_key_size
        self.kdist_own = kdist_own
        self.kdist_peer = kdist_peer

    @classmethod
    def from_c(cls, sec_params):
        return cls(bond=sec_params.bond,
                   mitm=sec_params.mitm,
                   le_sec_pairing=sec_params.lesc,
                   keypress_noti=sec_params.keypress,
                   io_caps=sec_params.io_caps,
                   oob=sec_params.oob,
                   min_key_size=sec_params.min_key_size,
                   max_key_size=sec_params.max_key_size,
                   kdist_own=BLEGapSecKeyDist.from_c(sec_params.kdist_own),
                   kdist_peer=BLEGapSecKeyDist.from_c(sec_params.kdist_peer))

    def to_c(self):
        sec_params = driver.ble_gap_sec_params_t()
        sec_params.bond = self.bond
        sec_params.mitm = self.mitm
        sec_params.lesc = self.le_sec_pairing
        sec_params.keypress = self.keypress_noti
        sec_params.io_caps = self.io_caps
        sec_params.oob = self.oob
        sec_params.min_key_size = self.min_key_size
        sec_params.max_key_size = self.max_key_size
        sec_params.kdist_own = self.kdist_own.to_c()
        sec_params.kdist_peer = self.kdist_peer.to_c()
        return sec_params

    def __repr__(self):
        return "{}(bond={!r}, mitm={!r}, lesc={!r}, " \
               "keypress_noti={!r}, io_caps={!r}, oob={!r}, " \
               "min_key_size={!r}, max_key_size={!r}, " \
               "kdist_own={!r}, kdist_peer={!r})".format(self.__class__.__name__, self.bond, self.mitm,
                                                         self.le_sec_pairing, self.keypress_noti, self.io_caps,
                                                         self.oob, self.min_key_size, self.max_key_size,
                                                         self.kdist_own, self.kdist_peer)


class BLEGapMasterId(object):
    RAND_LEN = driver.BLE_GAP_SEC_RAND_LEN

    def __init__(self, ediv=0, rand=""):
        self.ediv = ediv
        self.rand = rand

    def to_c(self):
        rand_array = util.list_to_uint8_array(self.rand)
        master_id = driver.ble_gap_master_id_t()
        master_id.ediv = self.ediv
        master_id.rand = rand_array.cast()
        return master_id

    @classmethod
    def from_c(cls, master_id):
        rand = bytearray(util.uint8_array_to_list(master_id.rand, cls.RAND_LEN))
        ediv = master_id.ediv
        return cls(ediv, rand)

    def __repr__(self):
        return "{}(e: {!r}, r: {!r})".format(self.__class__.__name__, self.ediv, binascii.hexlify(self.rand))


class BLEGapEncryptInfo(object):
    KEY_LENGTH = driver.BLE_GAP_SEC_KEY_LEN

    def __init__(self, ltk="", lesc=False, auth=False):
        self.ltk = ltk
        self.lesc = lesc
        self.auth = auth

    def to_c(self):
        info = driver.ble_gap_enc_info_t()
        info.ltk = util.list_to_uint8_array(self.ltk).cast()
        info.lesc = self.lesc
        info.auth = self.auth
        info.ltk_len = len(self.ltk)
        return info

    @classmethod
    def from_c(cls, info):
        ltk = bytearray(util.uint8_array_to_list(info.ltk, cls.KEY_LENGTH))
        lesc = info.lesc
        auth = info.auth
        return cls(ltk, lesc, auth)

    def __repr__(self):
        if not self.ltk:
            return ""
        return "Encrypt(ltk: {}, lesc: {}, auth: {})".format(binascii.hexlify(self.ltk), self.lesc, self.auth)


class BLEGapEncryptKey(object):
    def __init__(self, enc_info=None, master_id=None):
        self.enc_info = enc_info
        self.master_id = master_id

    def to_c(self):
        key = driver.ble_gap_enc_key_t()

        if self.enc_info:
            key.enc_info = self.enc_info.to_c()
        if self.master_id:
            key.master_id = self.master_id.to_c()
        return key

    @classmethod
    def from_c(cls, key):
        enc_info = BLEGapEncryptInfo.from_c(key.enc_info)
        master_id = BLEGapMasterId.from_c(key.master_id)
        return cls(enc_info, master_id)

    def __repr__(self):
        if not self.enc_info:
            return ""
        return "key: {}, master_id: {}".format(self.enc_info, self.master_id)


class BLEGapIdKey(object):
    KEY_LENGTH = driver.BLE_GAP_SEC_KEY_LEN

    def __init__(self, irk=b"", peer_addr=None):
        self.irk = irk
        self.peer_addr = peer_addr

    def to_c(self):
        irk_key = driver.ble_gap_id_key_t()

        irk = driver.ble_gap_irk_t()
        irk.irk = util.list_to_uint8_array(self.irk).cast()
        irk_key.id_info = irk

        if self.peer_addr:
            addr = self.peer_addr.to_c()
            irk_key.id_addr_info = addr
        return irk_key

    @classmethod
    def from_c(cls, id_key):
        irk = bytearray(util.uint8_array_to_list(id_key.id_info.irk, cls.KEY_LENGTH))
        addr = BLEGapAddr.from_c(id_key.id_addr_info)
        return cls(irk, addr)

    def __repr__(self):
        if not self.irk:
            return ""
        return "irk: {}, peer: {}".format(binascii.hexlify(self.key).decode("ascii"), self.peer_addr)


class BLEGapPublicKey(object):
    KEY_LENGTH = driver.BLE_GAP_LESC_P256_PK_LEN

    def __init__(self, key=b""):
        self.key = key

    def to_c(self):
        key = driver.ble_gap_lesc_p256_pk_t()
        key.pk = util.list_to_uint8_array(self.key).cast()
        return key

    @classmethod
    def from_c(cls, key):
        key_data = bytearray(util.uint8_array_to_list(key.pk, cls.KEY_LENGTH))
        return cls(key_data)

    def __repr__(self):
        if not self.key:
            return ""
        return binascii.hexlify(self.key).decode("ascii")


class BLEGapDhKey(object):
    KEY_LENGTH = driver.BLE_GAP_LESC_DHKEY_LEN

    def __init__(self, key=b""):
        self.key = key

    def to_c(self):
        key = driver.ble_gap_lesc_dhkey_t()
        key.key = util.list_to_uint8_array(self.key).cast()

        return key

    @classmethod
    def from_c(cls, key):
        key_data = bytearray(util.uint8_array_to_list(key.key, cls.KEY_LENGTH))
        return cls(key_data)

    def __repr__(self):
        if not self.key:
            return ""
        return binascii.hexlify(self.key).decode("ascii")


class BLEGapSignKey(object):
    KEY_LENGTH = driver.BLE_GAP_SEC_KEY_LEN

    def __init__(self, key=b""):
        self.key = key

    def to_c(self):
        key = driver.ble_gap_sign_info_t()
        key.csrk = util.list_to_uint8_array(self.key).cast()
        return key

    @classmethod
    def from_c(cls, key):
        key_data = bytearray(util.uint8_array_to_list(key.csrk, cls.KEY_LENGTH))
        return cls(key_data)

    def __repr__(self):
        if not self.key:
            return ""
        return binascii.hexlify(self.key).decode("ascii")


class BLEGapSecKeys(object):
    def __init__(self, enc_key=None, id_key=None, sign_key=None, public_key=None):
        if not enc_key:
            enc_key = BLEGapEncryptKey()
        if not id_key:
            id_key = BLEGapIdKey()
        if not sign_key:
            sign_key = BLEGapSignKey()
        if not public_key:
            public_key = BLEGapPublicKey()
        self.enc_key = enc_key        # type: BLEGapEncryptKey
        self.id_key = id_key          # type: BLEGapIdKey
        self.sign_key = sign_key      # type: BLEGapSignKey
        self.public_key = public_key  # type: BLEGapPublicKey

    def to_c(self):
        keys = driver.ble_gap_sec_keys_t()
        keys.p_enc_key = self.enc_key.to_c()
        keys.p_id_key = self.id_key.to_c()
        keys.p_sign_key = self.sign_key.to_c()
        keys.p_pk = self.public_key.to_c()
        return keys

    @classmethod
    def from_c(cls, keys):
        enc_key = BLEGapEncryptKey.from_c(keys.p_enc_key)
        id_key = BLEGapIdKey.from_c(keys.p_id_key)
        sign_key = BLEGapSignKey.from_c(keys.p_sign_key)
        pk = BLEGapPublicKey.from_c(keys.p_pk)
        return cls(enc_key, id_key, sign_key, pk)

    def __repr__(self):
        return "{}(enc: {}, id: {}, sign: {}, pk: {})".format(self.__class__.__name__, self.enc_key, self.id_key,
                                                              self.sign_key, self.public_key)


class BLEGapSecKeyset(object):
    def __init__(self, own_keys=None, peer_keys=None):
        if not own_keys:
            own_keys = BLEGapSecKeys()
        if not peer_keys:
            peer_keys = BLEGapSecKeys()
        self.own_keys = own_keys
        self.peer_keys = peer_keys
        self.ble_keyset = self.to_c()

    def to_c(self):
        self.ble_keyset = driver.ble_gap_sec_keyset_t()
        self.ble_keyset.keys_own = self.own_keys.to_c()
        self.ble_keyset.keys_peer = self.peer_keys.to_c()
        return self.ble_keyset

    def reload(self):
        self.own_keys = BLEGapSecKeys.from_c(self.ble_keyset.keys_own)
        self.peer_keys = BLEGapSecKeys.from_c(self.ble_keyset.keys_peer)

    @classmethod
    def from_c(cls, keyset):
        own_keys = BLEGapSecKeys.from_c(keyset.keys_own)
        peer_keys = BLEGapSecKeys.from_c(keyset.keys_peer)
        return cls(own_keys, peer_keys)

    def __repr__(self):
        return "{}(own: {!r}, peer: {!r})".format(self.__class__.__name__, self.own_keys, self.peer_keys)



from blatann.gatt import SecurityLevel
from blatann.services.glucose.constants import GLUCOSE_SERVICE_UUID
from blatann.services.glucose.data_types import *
from blatann.services.glucose.database import IGlucoseDatabase, BasicGlucoseDatabase

from blatann.services.glucose.service import GlucoseServer as _GlucoseServer


def add_glucose_service(gatts_database, glucose_database, security_level=SecurityLevel.OPEN,
                        include_context_characteristic=True):
    """
    Adds the Glucose bluetooth service to the Gatt Server database

    :param gatts_database: The GATT database to add the service to
    :type gatts_database: blatann.gatt.gatts.GattsDatabase
    :param glucose_database: The database which holds the glucose measurements
    :type glucose_database: IGlucoseDatabase
    :param security_level: The security level for the record-access control point of the service
    :type security_level: SecurityLevel
    :param include_context_characteristic: flag whether or not to include the
                                           optional context characteristic in the service. If this is False,
                                           any context stored with glucose measurements will not be reported.
    """
    return _GlucoseServer.add_to_database(gatts_database, glucose_database, security_level,
                                          include_context_characteristic)



"""
This example demonstrates using Bluetooth SIG's defined Current Time service as a peripheral.
"""
import binascii
import datetime
from blatann import BleDevice
from blatann.gap import advertising
from blatann.utils import setup_logger
from blatann.services import current_time
from blatann.waitables import GenericWaitable


logger = setup_logger(level="INFO")


def on_connect(peer, event_args):
    """
    Event callback for when a central device connects to us

    :param peer: The peer that connected to us
    :type peer: blatann.peer.Client
    :param event_args: None
    """
    if peer:
        logger.info("Connected to peer")
    else:
        logger.warning("Connection timed out")


def on_disconnect(peer, event_args):
    """
    Event callback for when the client disconnects from us (or when we disconnect from the client)

    :param peer: The peer that disconnected
    :type peer: blatann.peer.Client
    :param event_args: The event args
    :type event_args: blatann.event_args.DisconnectionEventArgs
    """
    logger.info("Disconnected from peer, reason: {}".format(event_args.reason))


def on_current_time_write(characteristic, event_args):
    """
    Callback registered to be triggered whenever the Current Time characteristic is written to

    :param characteristic:
    :param event_args: The write event args
    :type event_args: blatann.event_args.DecodedWriteEventArgs
    """
    # event_args.value is of type current_time.CurrentTime
    logger.info("Current time written to, new value: {}. "
                "Raw: {}".format(event_args.value, binascii.hexlify(event_args.raw_value)))


def on_local_time_info_write(characteristic, event_args):
    """
    Callback registered to be triggered whenever the Local Time Info characteristic is written to

    :param characteristic:
    :param event_args: The write event args
    :type event_args: blatann.event_args.DecodedWriteEventArgs
    """
    # event_args.value is of type current_time.LocalTimeInfo
    logger.info("Local Time info written to, new value: {}. "
                "Raw: {}".format(event_args.value, binascii.hexlify(event_args.raw_value)))


def main(serial_port):
    # Create and open the device
    ble_device = BleDevice(serial_port)
    ble_device.open()

    # Create and add the current time service to the database.
    # Tweak the flags below to change how the service is set up
    current_time_service = current_time.add_current_time_service(ble_device.database,
                                                                 enable_writes=True,
                                                                 enable_local_time_info=True,
                                                                 enable_reference_info=False)

    # Register handlers for when the characteristics are written to.
    # These will only be triggered if enable_writes above is true
    current_time_service.on_current_time_write.register(on_current_time_write)
    current_time_service.on_local_time_info_write.register(on_local_time_info_write)

    # Demo of the different ways to manually or automatically control the reported time

    # Example 1: Automatically reference system time.
    #            All logic is handled within the service and reports the time whenever the characteristic is read
    # Example 2: Manually report the time 1 day behind using callback method.
    #            Register a user-defined callback to retrieve the current time to report back to the client
    # Example 3: Manually report the time 1 hour ahead by setting the base time
    #            Set the characteristic's base time to 1 day ahead and allow the service to auto-increment from there
    example_mode = 1

    if example_mode == 1:
        # configure_automatic() also sets up the Local Time characteristic (if enabled)
        # to just set the automatic time and leave Local Time unconfigured, use set_time() with no parameters
        current_time_service.configure_automatic()
    elif example_mode == 2:
        def on_time_read():
            d = datetime.datetime.now() - datetime.timedelta(days=1)
            logger.info("Getting time: {}".format(d))
            return d
        current_time_service.set_time(characteristic_read_callback=on_time_read)
    elif example_mode == 3:
        base_time = datetime.datetime.now() + datetime.timedelta(hours=1)
        current_time_service.set_time(base_time)

    # Register listeners for when the client connects and disconnects
    ble_device.client.on_connect.register(on_connect)
    ble_device.client.on_disconnect.register(on_disconnect)

    # Advertise the Current Time service
    adv_data = advertising.AdvertisingData(local_name="Current Time", flags=0x06,
                                           service_uuid16s=current_time.CURRENT_TIME_SERVICE_UUID)
    ble_device.advertiser.set_advertise_data(adv_data)

    logger.info("Advertising")
    ble_device.advertiser.start(timeout_sec=0, auto_restart=True)

    # Create a waitable that waits 5 minutes, then exits
    w = GenericWaitable()
    w.wait(5 * 60, exception_on_timeout=False)

    logger.info("Done")
    ble_device.close()


if __name__ == '__main__':
    main("COM8")



import logging
from blatann.services.glucose.constants import *
from blatann.services.glucose.data_types import *
from blatann.services.glucose.racp import *
from blatann.services.glucose.database import IGlucoseDatabase
from blatann.gatt import SecurityLevel
from blatann.gatt.gatts import GattsService, GattsCharacteristicProperties


logger = logging.getLogger(__name__)


class GlucoseServer(object):
    def __init__(self, service, glucose_database, security_level=SecurityLevel.OPEN,
                 include_context_characteristic=True):
        """
        :type service: GattsService
        :type glucose_database: IGlucoseDatabase
        :param security_level:
        """
        self.service = service
        self.database = glucose_database

        measurement_char_props = GattsCharacteristicProperties(read=False, notify=True, max_length=20)
        feature_props = GattsCharacteristicProperties(read=True, max_length=GlucoseFeatures.encoded_size(),
                                                      variable_length=False)
        racp_props = GattsCharacteristicProperties(read=False, write=True, indicate=True, max_length=20,
                                                   variable_length=True, security_level=security_level)
        self.measurement_characteristic = service.add_characteristic(MEASUREMENT_CHARACTERISTIC_UUID, measurement_char_props)
        if include_context_characteristic:
            self.context_characteristic = service.add_characteristic(MEASUREMENT_CONTEXT_CHARACTERISTIC_UUID,
                                                                     measurement_char_props)
            self.context_characteristic.on_notify_complete.register(self._on_notify_complete)
        else:
            self.context_characteristic = None
        self.feature_characteristic = service.add_characteristic(FEATURE_CHARACTERISTIC_UUID, feature_props)
        self.racp_characteristic = service.add_characteristic(RACP_CHARACTERISTIC_UUID, racp_props)
        self.racp_characteristic.on_write.register(self._on_racp_write)
        self._current_command = None
        self._records_to_report = []  # type: list[GlucoseMeasurement]
        self._active_notifications = []
        self.service.peer.on_disconnect.register(self._on_disconnect)
        self.measurement_characteristic.on_notify_complete.register(self._on_notify_complete)

    def set_features(self, features):
        """
        Sets the features for the Glucose Feature characteristic

        :param features: The supported features of the sensor
        :type features: GlucoseFeatures
        """
        self.feature_characteristic.set_value(features.encode().value, False)

    def _report_next_record(self):
        if self._current_command:
            if len(self._records_to_report) > 0:
                next_record = self._records_to_report.pop(0)
                noti_id = self.measurement_characteristic.notify(next_record.encode().value).id
                self._active_notifications.append(noti_id)
                if next_record.context and self.context_characteristic and self.context_characteristic.client_subscribed:
                    noti_id = self.context_characteristic.notify(next_record.context.encode().value).id
                    self._active_notifications.append(noti_id)
            else:
                # Done reporting
                response = RacpResponse(self._current_command.opcode, RacpResponseCode.success)
                self._current_command = None
                self.racp_characteristic.notify(response.encode().value)
        return None

    def _on_report_records_request(self, command):
        """
        :type command: RacpCommand
        :return:
        """
        if self._current_command is not None:
            return RacpResponseCode.procedure_not_completed
        if command.filter_type not in [None, FilterType.sequence_number]:
            return RacpResponseCode.operand_not_supported

        if command.operator in [RacpOperator.first_record, RacpOperator.last_record]:
            if command.operator == RacpOperator.first_record:
                record = self.database.first_record()
            else:
                record = self.database.last_record()
            records = [record]
        elif command.operator in [RacpOperator.all_records, RacpOperator.less_than_or_equal_to,
                                  RacpOperator.greater_than_or_equal_to, RacpOperator.within_range_inclusive]:
            min_seq, max_seq = command.get_filter_min_max()
            records = self.database.get_records(min_seq, max_seq)
        else:
            return RacpResponseCode.invalid_operator

        if len(records) == 0:
            return RacpResponseCode.no_records_found

        # Start reporting records
        self._records_to_report = sorted(records, key=lambda r: r.sequence_number)
        self._current_command = command

        self._report_next_record()
        # Do not send a response, will be sent once all records reported
        return None

    def _on_report_num_records_request(self, command):
        min_seq, max_seq = command.get_filter_min_max()
        return self.database.record_count(min_seq, max_seq)

    def _on_delete_records_request(self, command):
        min_seq, max_seq = command.get_filter_min_max()
        return self.database.delete_records(min_seq, max_seq)

    def _on_abort_operation(self):
        if self._current_command is not None:
            self._current_command = None
            self._records_to_report = []
            return RacpResponseCode.success
        return RacpResponseCode.abort_not_successful

    def _on_racp_write(self, characteristic, event_args):
        """
        :param characteristic:
        :type event_args: blatann.event_args.WriteEventArgs
        """
        stream = ble_data_types.BleDataStream(event_args.value)
        command = RacpCommand.decode(stream)

        response = None

        if command.opcode == RacpOpcode.report_number_of_records:
            record_count = self._on_report_num_records_request(command)
            response = RacpResponse(record_count=record_count)
        else:
            response_code = None
            if command.opcode == RacpOpcode.report_stored_records:
                response_code = self._on_report_records_request(command)
            elif command.opcode == RacpOpcode.delete_stored_records:
                response_code = self._on_delete_records_request(command)
            elif command.opcode == RacpOpcode.abort_operation:
                response_code = self._on_abort_operation()

            if response_code is not None:
                response = RacpResponse(command.opcode, response_code)

        if response:
            self.racp_characteristic.notify(response.encode().value)

    def _on_notify_complete(self, characteristic, event_args):
        """
        :param characteristic:
        :type event_args: blatann.event_args.NotificationCompleteEventArgs
        """
        if event_args.id in self._active_notifications:
            self._active_notifications.remove(event_args.id)

        if self._current_command is not None:
            self._report_next_record()

    def _on_disconnect(self, peer, event_args):
        self._current_command = None
        self._active_notifications = []
        self._records_to_report = []

    @classmethod
    def add_to_database(cls, gatts_database, glucose_database, security_level=SecurityLevel.OPEN,
                        include_context_characteristic=True):
        """
        :type gatts_database: blatann.gatt.gatts.GattsDatabase
        :type glucose_database: IGlucoseDatabase
        :type security_level: SecurityLevel
        :type include_context_characteristic: bool
        """
        service = gatts_database.add_service(GLUCOSE_SERVICE_UUID)
        return GlucoseServer(service, glucose_database, security_level, include_context_characteristic)



import logging
import sys
from blatann.utils import _threading


LOG_FORMAT = "[%(asctime)s] [%(threadName)s] [%(name)s.%(funcName)s:%(lineno)s] [%(levelname)s]: %(message)s"


def setup_logger(name=None, level="DEBUG"):
    logger = logging.getLogger(name)
    logger.setLevel(level)
    formatter = logging.Formatter(LOG_FORMAT)
    stream_handler = logging.StreamHandler(sys.stderr)
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)
    return logger


import logging
import threading
import enum

from blatann.gap import smp_crypto
from blatann.gap.bond_db import BondingData
from blatann.nrf import nrf_types, nrf_events
from blatann.exceptions import InvalidStateException, InvalidOperationException
from blatann.event_type import EventSource, Event
from blatann.waitables.event_waitable import EventWaitable
from blatann.event_args import *


logger = logging.getLogger(__name__)

# Enum used to report IO Capabilities for the pairing process
IoCapabilities = nrf_types.BLEGapIoCaps

# Enum of the status codes emitted during security procedures
SecurityStatus = nrf_types.BLEGapSecStatus

# Enum of the different Pairing passkeys to be entered by the user (passcode, out-of-band, etc.)
AuthenticationKeyType = nrf_types.BLEGapAuthKeyType


class SecurityLevel(enum.Enum):
    """
    Security levels used for defining GATT server characteristics
    """
    NO_ACCESS = 0
    OPEN = 1
    JUST_WORKS = 2
    MITM = 3
    LESC_MITM = 4


class SecurityParameters(object):
    """
    Class representing the desired security parameters for a given connection
    """
    def __init__(self, passcode_pairing=False, io_capabilities=IoCapabilities.KEYBOARD_DISPLAY,
                 bond=False, out_of_band=False, reject_pairing_requests=False, lesc_pairing=False):
        self.passcode_pairing = passcode_pairing
        self.io_capabilities = io_capabilities
        self.bond = bond
        self.out_of_band = out_of_band
        self.reject_pairing_requests = reject_pairing_requests
        self.lesc_pairing = lesc_pairing


class SecurityManager(object):
    """
    Handles performing security procedures with a connected peer
    """
    def __init__(self, ble_device, peer, security_parameters):
        """
        :type ble_device: blatann.BleDevice
        :type peer: blatann.peer.Peer
        :type security_parameters: SecurityParameters
        """
        self.ble_device = ble_device
        self.peer = peer
        self.security_params = security_parameters
        self._pairing_in_process = False
        self._initiated_encryption = False
        self._is_previously_bonded_device = False
        self._on_authentication_complete_event = EventSource("On Authentication Complete", logger)
        self._on_passkey_display_event = EventSource("On Passkey Display", logger)
        self._on_passkey_entry_event = EventSource("On Passkey Entry", logger)
        self._on_security_level_changed_event = EventSource("Security Level Changed", logger)
        self.peer.on_connect.register(self._on_peer_connected)
        self._auth_key_resolve_thread = threading.Thread()
        self.keyset = nrf_types.BLEGapSecKeyset()
        self.bond_db_entry = None
        self._security_level = SecurityLevel.NO_ACCESS
        self._private_key = smp_crypto.lesc_generate_private_key()
        self._public_key = self._private_key.public_key()
        self.keyset.own_keys.public_key.key = smp_crypto.lesc_pubkey_to_raw(self._public_key)

    """
    Events
    """

    @property
    def on_pairing_complete(self):
        """
        Event that is triggered when pairing completes with the peer
        EventArgs type: PairingCompleteEventArgs

        :return: an Event which can have handlers registered to and deregistered from
        :rtype: Event
        """
        return self._on_authentication_complete_event

    @property
    def on_security_level_changed(self):
        """
        Event that is triggered when the security/encryption level changes. This can be triggered from
        a pairing sequence or if a bonded client starts the encryption handshaking using the stored LTKs.

        Note: This event is triggered before on_pairing_complete

        EventArgs type: SecurityLevelChangedEventArgs
        :return: an Event which can have handlers registered to and deregestestered from
        :rtype: Event
        """
        return self._on_security_level_changed_event

    @property
    def on_passkey_display_required(self):
        """
        Event that is triggered when a passkey needs to be displayed to the user
        EventArgs type: PasskeyDisplayEventArgs

        :return: an Event which can have handlers registered to and deregistered from
        :rtype: Event
        """
        return self._on_passkey_display_event

    @property
    def on_passkey_required(self):
        """
        Event that is triggered when a passkey needs to be entered by the user
        EventArgs type: PasskeyEntryEventArgs

        :return: an Event which can have handlers registered to and deregistered from
        :rtype: Event
        """
        return self._on_passkey_entry_event

    @property
    def is_previously_bonded(self):
        """
        Gets if the peer this security manager is for was bonded in a previous connection

        :return: True if previously bonded, False if not
        """
        return self._is_previously_bonded_device

    @property
    def security_level(self):
        """
        Gets the current security level of the connection

        :rtype: SecurityLevel
        """
        return self._security_level

    """
    Public Methods
    """

    def set_security_params(self, passcode_pairing, io_capabilities, bond, out_of_band, reject_pairing_requests=False,
                            lesc_pairing=False):
        """
        Sets the security parameters to use with the peer

        :param passcode_pairing: Flag indicating that passcode pairing is required
        :type passcode_pairing: bool
        :param io_capabilities: The input/output capabilities of this device
        :type io_capabilities: IoCapabilities
        :param bond: Flag indicating that long-term bonding should be performed
        :type bond: bool
        :param out_of_band: Flag indicating if out-of-band pairing is supported
        :type out_of_band: bool
        :param reject_pairing_requests: Flag indicating that all security requests by the peer should be rejected
        :type reject_pairing_requests: bool
        :param lesc_pairing: Flag indicating that LE Secure Pairing methods are supported
        """
        self.security_params = SecurityParameters(passcode_pairing, io_capabilities, bond, out_of_band,
                                                  reject_pairing_requests, lesc_pairing)

    def pair(self, force_repairing=False):
        """
        Starts the encrypting process with the peer. If the peer has already been bonded to,
        Starts the pairing process with the peer given the set security parameters
        and returns a Waitable which will fire when the pairing process completes, whether successful or not.
        Waitable returns two parameters: (Peer, PairingCompleteEventArgs)

        :return: A waitiable that will fire when pairing is complete
        :rtype: blatann.waitables.EventWaitable
        """
        if self._pairing_in_process or self._initiated_encryption:
            raise InvalidStateException("Security manager busy")
        if self.security_params.reject_pairing_requests:
            raise InvalidOperationException("Cannot initiate pairing while rejecting pairing requests")

        # if in the client role and don't want to force a re-pair, check for bonding data first
        if self.peer.is_peripheral and not force_repairing:
            bond_entry = self._find_db_entry(self.peer.peer_address)
            if bond_entry:
                logger.info("Re-establishing encryption with peer using LTKs")
                self.ble_device.ble_driver.ble_gap_encrypt(self.peer.conn_handle,
                                                           bond_entry.bonding_data.own_ltk.master_id,
                                                           bond_entry.bonding_data.own_ltk.enc_info)
                self._initiated_encryption = True

        sec_params = self._get_security_params()
        self.ble_device.ble_driver.ble_gap_authenticate(self.peer.conn_handle, sec_params)
        self._pairing_in_process = True
        return EventWaitable(self.on_pairing_complete)

    def use_debug_lesc_key(self):
        """
        Changes the security settings to use the debug public/private key-pair for future LESC pairing interactions.
        The key is defined in the Core Bluetooth Specification v4.2 Vol.3, Part H, Section 2.3.5.6.

        .. warning:: Using this key allows Bluetooth sniffers to be able to decode the encrypted traffic over the air
        """
        self._private_key = smp_crypto.LESC_DEBUG_PRIVATE_KEY
        self._public_key = smp_crypto.LESC_DEBUG_PUBLIC_KEY
        self.keyset.own_keys.public_key.key = smp_crypto.lesc_pubkey_to_raw(self._public_key)

    """
    Private Methods
    """

    def _on_peer_connected(self, peer, event_args):
        # Reset the
        self._pairing_in_process = False
        self._initiated_encryption = False
        self._security_level = SecurityLevel.OPEN
        self.peer.driver_event_subscribe(self._on_security_params_request, nrf_events.GapEvtSecParamsRequest)
        self.peer.driver_event_subscribe(self._on_authentication_status, nrf_events.GapEvtAuthStatus)
        self.peer.driver_event_subscribe(self._on_conn_sec_status, nrf_events.GapEvtConnSecUpdate)
        self.peer.driver_event_subscribe(self._on_auth_key_request, nrf_events.GapEvtAuthKeyRequest)
        self.peer.driver_event_subscribe(self._on_passkey_display, nrf_events.GapEvtPasskeyDisplay)
        self.peer.driver_event_subscribe(self._on_security_info_request, nrf_events.GapEvtSecInfoRequest)
        self.peer.driver_event_subscribe(self._on_lesc_dhkey_request, nrf_events.GapEvtLescDhKeyRequest)
        # Search the bonding DB for this peer's info
        self.bond_db_entry = self._find_db_entry(self.peer.peer_address)
        if self.bond_db_entry:
            logger.info("Connected to previously bonded device {}".format(self.bond_db_entry.peer_addr))
            self._is_previously_bonded_device = True

    def _find_db_entry(self, peer_address):
        if peer_address.addr_type == nrf_types.BLEGapAddrTypes.random_private_non_resolvable:
            return None

        for r in self.ble_device.bond_db:
            if self.peer.is_client != r.peer_is_client:
                continue

            # If peer address is public or random static, check directly if they match (no IRK needed)
            if peer_address.addr_type in [nrf_types.BLEGapAddrTypes.random_static, nrf_types.BLEGapAddrTypes.public]:
                if r.peer_addr == peer_address:
                    return r
            elif smp_crypto.private_address_resolves(peer_address, r.bonding_data.peer_id.irk):
                logger.info("Resolved Peer ID to {}".format(r.peer_addr))
                return r

        return None

    def _get_security_params(self):
        keyset_own = nrf_types.BLEGapSecKeyDist(True, True, False, False)
        keyset_peer = nrf_types.BLEGapSecKeyDist(True, True, False, False)
        sec_params = nrf_types.BLEGapSecParams(self.security_params.bond, self.security_params.passcode_pairing,
                                               self.security_params.lesc_pairing, False,
                                               self.security_params.io_capabilities,
                                               self.security_params.out_of_band, 7, 16, keyset_own, keyset_peer)
        return sec_params

    def _on_security_params_request(self, driver, event):
        """
        :type event: nrf_events.GapEvtSecParamsRequest
        """
        # Security parameters are only provided for clients
        sec_params = self._get_security_params() if self.peer.is_client else None

        if self.security_params.reject_pairing_requests:
            status = nrf_types.BLEGapSecStatus.pairing_not_supp
        else:
            status = nrf_types.BLEGapSecStatus.success

        self.ble_device.ble_driver.ble_gap_sec_params_reply(event.conn_handle, status, sec_params, self.keyset)

        if not self.security_params.reject_pairing_requests:
            self._pairing_in_process = True

    def _on_security_info_request(self, driver, event):
        """
        :type event: nrf_events.GapEvtSecInfoRequest
        """
        found_record = None
        # Find the database entry based on the sec info given
        for r in self.ble_device.bond_db:
            # Check that roles match
            if r.peer_is_client != self.peer.is_client:
                continue
            own_mid = r.bonding_data.own_ltk.master_id
            peer_mid = r.bonding_data.peer_ltk.master_id
            if event.master_id.ediv == own_mid.ediv and event.master_id.rand == own_mid.rand:
                logger.info("Found matching record with own master ID for sec info request")
                found_record = r
                break
            if event.master_id.ediv == peer_mid.ediv and event.master_id.rand == peer_mid.rand:
                logger.info("Found matching record with peer master ID for sec info request")
                found_record = r
                break

        if not found_record:
            logger.info("Unable to find Bonding record for peer master id {}".format(event.master_id))
            self.ble_device.ble_driver.ble_gap_sec_info_reply(event.conn_handle)
        else:
            self.bond_db_entry = found_record
            ltk = found_record.bonding_data.own_ltk
            id_key = found_record.bonding_data.peer_id
            self.ble_device.ble_driver.ble_gap_sec_info_reply(event.conn_handle, ltk.enc_info, id_key, None)

    def _on_lesc_dhkey_request(self, driver, event):
        """
        :type event: nrf_events.GapEvtLescDhKeyRequest
        """
        peer_public_key = smp_crypto.lesc_pubkey_from_raw(event.remote_public_key.key)
        dh_key = smp_crypto.lesc_compute_dh_key(self._private_key, peer_public_key, little_endian=True)
        self.ble_device.ble_driver.ble_gap_lesc_dhkey_reply(event.conn_handle, nrf_types.BLEGapDhKey(dh_key))

    def _on_conn_sec_status(self, driver, event):
        """
        :type event: nrf_events.GapEvtConnSecUpdate
        """
        self._security_level = SecurityLevel(event.sec_level)
        self._on_security_level_changed_event.notify(self.peer, SecurityLevelChangedEventArgs(self._security_level))
        if self._initiated_encryption:
            self._initiated_encryption = False
            if event.sec_level > 0 and event.sec_mode > 0:
                status = SecurityStatus.success
            else:
                # Peer failed to find/load the keys, return failure status code
                status = SecurityStatus.unspecified
            self._on_authentication_complete_event.notify(self.peer, PairingCompleteEventArgs(status, self.security_level))

    def _on_authentication_status(self, driver, event):
        """
        :type event: nrf_events.GapEvtAuthStatus
        """
        self._pairing_in_process = False
        self._on_authentication_complete_event.notify(self.peer, PairingCompleteEventArgs(event.auth_status,
                                                                                          self.security_level))
        # Save keys in the database if authenticated+bonded successfullly
        if event.auth_status == SecurityStatus.success and event.bonded:
            # Reload the keys from the C Memory space (were updated during the pairing process)
            self.keyset.reload()

            # If there wasn't a bond record initially, try again a second time using the new public peer address
            if not self.bond_db_entry:
                self.bond_db_entry = self._find_db_entry(self.keyset.peer_keys.id_key.peer_addr)

            # Still no bond DB entry, create a new one
            if not self.bond_db_entry:
                logger.info("New bonded device, creating a DB Entry")
                self.bond_db_entry = self.ble_device.bond_db.create()
                self.bond_db_entry.peer_is_client = self.peer.is_client
                self.bond_db_entry.peer_addr = self.keyset.peer_keys.id_key.peer_addr
                self.bond_db_entry.bonding_data = BondingData(self.keyset)
                self.ble_device.bond_db.add(self.bond_db_entry)
            else:  # update the bonding info
                logger.info("Updating bond key for peer {}".format(self.keyset.peer_keys.id_key.peer_addr))
                self.bond_db_entry.bonding_data = BondingData(self.keyset)

            # TODO: This doesn't belong here..
            self.ble_device.bond_db_loader.save(self.ble_device.bond_db)

    def _on_passkey_display(self, driver, event):
        """
        :type event: nrf_events.GapEvtPasskeyDisplay
        """
        def match_confirm(keys_match):
            if not self._pairing_in_process:
                return
            if keys_match:
                key_type = nrf_types.BLEGapAuthKeyType.PASSKEY
            else:
                key_type = nrf_types.BLEGapAuthKeyType.NONE

            self.ble_device.ble_driver.ble_gap_auth_key_reply(event.conn_handle, key_type, None)

        event_args = PasskeyDisplayEventArgs(event.passkey, event.match_request, match_confirm)
        if event.match_request:
            self._auth_key_resolve_thread = threading.Thread(name="{} Passkey Confirm".format(self.peer.conn_handle),
                                                             target=self._on_passkey_display_event.notify,
                                                             args=(self.peer, event_args))
            self._auth_key_resolve_thread.daemon = True
            self._auth_key_resolve_thread.start()
        else:
            self._on_passkey_display_event.notify(self.peer, event_args)

    def _on_auth_key_request(self, driver, event):
        """
        :type event: nrf_events.GapEvtAuthKeyRequest
        """
        def resolve(passkey):
            if not self._pairing_in_process:
                return
            if isinstance(passkey, int):
                passkey = "{:06d}".format(passkey).encode("ascii")
            elif isinstance(passkey, str):
                passkey = passkey.encode("ascii")
            self.ble_device.ble_driver.ble_gap_auth_key_reply(self.peer.conn_handle, event.key_type, passkey)

        self._auth_key_resolve_thread = threading.Thread(name="{} Passkey Entry".format(self.peer.conn_handle),
                                                         target=self._on_passkey_entry_event.notify,
                                                         args=(self.peer, PasskeyEntryEventArgs(event.key_type, resolve)))
        self._auth_key_resolve_thread.daemon = True
        self._auth_key_resolve_thread.start()

    def _on_timeout(self, driver, event):
        """
        :type event: nrf_events.GapEvtTimeout
        """
        if event.src != nrf_types.BLEGapTimeoutSrc.security_req:
            return
        self._on_authentication_complete_event.notify(self.peer, PairingCompleteEventArgs(SecurityStatus.timeout,
                                                                                          self.security_level))



from typing import TypeVar, Generic, Tuple
from blatann.waitables.waitable import Waitable


TSender = TypeVar("TSender")
TEvent = TypeVar("TEvent")


class EventWaitable(Waitable, Generic[TSender, TEvent]):
    def __init__(self, event):
        """
        :type event: blatann.event_type.Event
        """
        super(EventWaitable, self).__init__()
        self._event = event
        self._event.register(self._on_event)

    def _on_event(self, *args):
        self._event.deregister(self._on_event)
        self._notify(*args)

    def _on_timeout(self):
        self._event.deregister(self._on_event)

    def wait(self, timeout=None, exception_on_timeout=True) -> Tuple[TSender, TEvent]:
        res = super(EventWaitable, self).wait(timeout, exception_on_timeout)
        if res is None:  # Timeout, send None, None for the sender and event_args
            return None, None
        return res


class IdBasedEventWaitable(EventWaitable):
    def __init__(self, event, event_id):
        super(IdBasedEventWaitable, self).__init__(event)
        self.id = event_id

    def _on_event(self, sender, event_args):
        if event_args.id == self.id:
            super(IdBasedEventWaitable, self)._on_event(sender, event_args)



from enum import Enum, IntEnum
import logging
from pc_ble_driver_py.exceptions import NordicSemiException
from blatann.nrf.nrf_dll_load import driver
import blatann.nrf.nrf_driver_types as util
from blatann.nrf.nrf_types.enums import *


logger = logging.getLogger(__name__)


class BLEGapAdvParams(object):
    def __init__(self, interval_ms, timeout_s, advertising_type=BLEGapAdvType.connectable_undirected):
        self.interval_ms = interval_ms
        self.timeout_s = timeout_s
        self.advertising_type = advertising_type

    def to_c(self):
        adv_params = driver.ble_gap_adv_params_t()
        adv_params.type = self.advertising_type.value
        adv_params.p_peer_addr = None  # Undirected advertisement.
        adv_params.fp = driver.BLE_GAP_ADV_FP_ANY
        adv_params.p_whitelist = None
        adv_params.interval = util.msec_to_units(self.interval_ms,
                                                 util.UNIT_0_625_MS)
        adv_params.timeout = self.timeout_s

        return adv_params

    def __repr__(self):
        return "{!r}(type: {!r}, interval: {!r}ms, timeout: {!r}s)".format(self.__class__.__name__,
                                                                           self.advertising_type,
                                                                           self.interval_ms,
                                                                           self.timeout_s)


class BLEGapScanParams(object):
    def __init__(self, interval_ms, window_ms, timeout_s):
        self.interval_ms = interval_ms
        self.window_ms = window_ms
        self.timeout_s = timeout_s

    def to_c(self):
        scan_params = driver.ble_gap_scan_params_t()
        scan_params.active = True
        scan_params.selective = False
        scan_params.p_whitelist = None
        scan_params.interval = util.msec_to_units(self.interval_ms,
                                                  util.UNIT_0_625_MS)
        scan_params.window = util.msec_to_units(self.window_ms,
                                                util.UNIT_0_625_MS)
        scan_params.timeout = self.timeout_s

        return scan_params


class BLEGapConnParams(object):
    def __init__(self, min_conn_interval_ms, max_conn_interval_ms, conn_sup_timeout_ms, slave_latency):
        self.min_conn_interval_ms = min_conn_interval_ms
        self.max_conn_interval_ms = max_conn_interval_ms
        self.conn_sup_timeout_ms = conn_sup_timeout_ms
        self.slave_latency = slave_latency

    @classmethod
    def from_c(cls, conn_params):
        return cls(min_conn_interval_ms=util.units_to_msec(conn_params.min_conn_interval,
                                                           util.UNIT_1_25_MS),
                   max_conn_interval_ms=util.units_to_msec(conn_params.max_conn_interval,
                                                           util.UNIT_1_25_MS),
                   conn_sup_timeout_ms=util.units_to_msec(conn_params.conn_sup_timeout,
                                                          util.UNIT_10_MS),
                   slave_latency=conn_params.slave_latency)

    def to_c(self):
        conn_params = driver.ble_gap_conn_params_t()
        conn_params.min_conn_interval = util.msec_to_units(self.min_conn_interval_ms,
                                                           util.UNIT_1_25_MS)
        conn_params.max_conn_interval = util.msec_to_units(self.max_conn_interval_ms,
                                                           util.UNIT_1_25_MS)
        conn_params.conn_sup_timeout = util.msec_to_units(self.conn_sup_timeout_ms,
                                                          util.UNIT_10_MS)
        conn_params.slave_latency = self.slave_latency

        return conn_params

    def __str__(self):
        return "{}(interval: [{!r}-{!r}] ms, timeout: {!r} ms, latency: {!r})".format(self.__class__.__name__,
                                                                                      self.min_conn_interval_ms,
                                                                                      self.max_conn_interval_ms,
                                                                                      self.conn_sup_timeout_ms,
                                                                                      self.slave_latency)


class BLEGapAddrTypes(IntEnum):
    public = int(driver.BLE_GAP_ADDR_TYPE_PUBLIC)
    random_static = int(driver.BLE_GAP_ADDR_TYPE_RANDOM_STATIC)
    random_private_resolvable = int(driver.BLE_GAP_ADDR_TYPE_RANDOM_PRIVATE_RESOLVABLE)
    random_private_non_resolvable = int(driver.BLE_GAP_ADDR_TYPE_RANDOM_PRIVATE_NON_RESOLVABLE)


class BLEGapAddr(object):

    def __init__(self, addr_type, addr):
        assert isinstance(addr_type, BLEGapAddrTypes), 'Invalid argument type'
        self.addr_type = addr_type
        self.addr = addr

    @classmethod
    def from_c(cls, addr):
        addr_list = util.uint8_array_to_list(addr.addr, driver.BLE_GAP_ADDR_LEN)
        addr_list.reverse()
        return cls(addr_type=BLEGapAddrTypes(addr.addr_type),
                   addr=addr_list)

    @classmethod
    def from_string(cls, addr_string):
        addr, addr_flag = addr_string.split(',')
        addr_list = [int(i, 16) for i in addr.split(':')]

        if addr_flag in ['p', 'public']:
            addr_type = BLEGapAddrTypes.public
        elif (addr_list[0] & 0b11000000) == 0b00000000:
            addr_type = BLEGapAddrTypes.random_private_non_resolvable
        elif (addr_list[0] & 0b11000000) == 0b01000000:
            addr_type = BLEGapAddrTypes.random_private_resolvable
        elif (addr_list[0] & 0b11000000) == 0b11000000:
            addr_type = BLEGapAddrTypes.random_static
        else:
            raise ValueError("Provided random address do not follow rules")  # TODO: Improve error message

        return cls(addr_type, addr_list)

    def to_c(self):
        addr_array = util.list_to_uint8_array(self.addr[::-1])
        addr = driver.ble_gap_addr_t()
        addr.addr_type = self.addr_type.value
        addr.addr = addr_array.cast()
        return addr

    def get_addr_type_str(self):
        if self.addr_type == BLEGapAddrTypes.public:
            return 'public'
        elif self.addr_type == BLEGapAddrTypes.random_private_non_resolvable:
            return 'nonres'
        elif self.addr_type == BLEGapAddrTypes.random_private_resolvable:
            return 'res'
        elif self.addr_type == BLEGapAddrTypes.random_static:
            return 'static'
        else:
            return 'err'

    def __eq__(self, other):
        if not isinstance(other, BLEGapAddr):
            other = BLEGapAddr.from_string(str(other))
        return str(self) == str(other)

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        value = 0
        for i, val in enumerate(self.addr[::-1]):
            value |= val << i*8
        return value

    def get_addr_flag(self):
        return {
            BLEGapAddrTypes.public: "p",
            BLEGapAddrTypes.random_static: "s",
            BLEGapAddrTypes.random_private_resolvable: "r",
            BLEGapAddrTypes.random_private_non_resolvable: "n"
        }[self.addr_type]

    def __str__(self):
        return '{},{}'.format(':'.join(['%02X' % i for i in self.addr]), self.get_addr_flag())

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, str(self))


class BLEAdvData(object):
    class Types(Enum):
        flags = driver.BLE_GAP_AD_TYPE_FLAGS
        service_16bit_uuid_more_available = driver.BLE_GAP_AD_TYPE_16BIT_SERVICE_UUID_MORE_AVAILABLE
        service_16bit_uuid_complete = driver.BLE_GAP_AD_TYPE_16BIT_SERVICE_UUID_COMPLETE
        service_32bit_uuid_more_available = driver.BLE_GAP_AD_TYPE_32BIT_SERVICE_UUID_MORE_AVAILABLE
        service_32bit_uuid_complete = driver.BLE_GAP_AD_TYPE_32BIT_SERVICE_UUID_COMPLETE
        service_128bit_uuid_more_available = driver.BLE_GAP_AD_TYPE_128BIT_SERVICE_UUID_MORE_AVAILABLE
        service_128bit_uuid_complete = driver.BLE_GAP_AD_TYPE_128BIT_SERVICE_UUID_COMPLETE
        short_local_name = driver.BLE_GAP_AD_TYPE_SHORT_LOCAL_NAME
        complete_local_name = driver.BLE_GAP_AD_TYPE_COMPLETE_LOCAL_NAME
        tx_power_level = driver.BLE_GAP_AD_TYPE_TX_POWER_LEVEL
        class_of_device = driver.BLE_GAP_AD_TYPE_CLASS_OF_DEVICE
        simple_pairing_hash_c = driver.BLE_GAP_AD_TYPE_SIMPLE_PAIRING_HASH_C
        simple_pairing_randimizer_r = driver.BLE_GAP_AD_TYPE_SIMPLE_PAIRING_RANDOMIZER_R
        security_manager_tk_value = driver.BLE_GAP_AD_TYPE_SECURITY_MANAGER_TK_VALUE
        security_manager_oob_flags = driver.BLE_GAP_AD_TYPE_SECURITY_MANAGER_OOB_FLAGS
        slave_connection_interval_range = driver.BLE_GAP_AD_TYPE_SLAVE_CONNECTION_INTERVAL_RANGE
        solicited_sevice_uuids_16bit = driver.BLE_GAP_AD_TYPE_SOLICITED_SERVICE_UUIDS_16BIT
        solicited_sevice_uuids_128bit = driver.BLE_GAP_AD_TYPE_SOLICITED_SERVICE_UUIDS_128BIT
        service_data = driver.BLE_GAP_AD_TYPE_SERVICE_DATA
        public_target_address = driver.BLE_GAP_AD_TYPE_PUBLIC_TARGET_ADDRESS
        random_target_address = driver.BLE_GAP_AD_TYPE_RANDOM_TARGET_ADDRESS
        appearance = driver.BLE_GAP_AD_TYPE_APPEARANCE
        advertising_interval = driver.BLE_GAP_AD_TYPE_ADVERTISING_INTERVAL
        le_bluetooth_device_address = driver.BLE_GAP_AD_TYPE_LE_BLUETOOTH_DEVICE_ADDRESS
        le_role = driver.BLE_GAP_AD_TYPE_LE_ROLE
        simple_pairng_hash_c256 = driver.BLE_GAP_AD_TYPE_SIMPLE_PAIRING_HASH_C256
        simple_pairng_randomizer_r256 = driver.BLE_GAP_AD_TYPE_SIMPLE_PAIRING_RANDOMIZER_R256
        service_data_32bit_uuid = driver.BLE_GAP_AD_TYPE_SERVICE_DATA_32BIT_UUID
        service_data_128bit_uuid = driver.BLE_GAP_AD_TYPE_SERVICE_DATA_128BIT_UUID
        uri = driver.BLE_GAP_AD_TYPE_URI
        information_3d_data = driver.BLE_GAP_AD_TYPE_3D_INFORMATION_DATA
        manufacturer_specific_data = driver.BLE_GAP_AD_TYPE_MANUFACTURER_SPECIFIC_DATA

    def __init__(self, **kwargs):
        self.records = dict()
        for k in kwargs:
            self.records[BLEAdvData.Types[k]] = kwargs[k]

    def to_list(self):
        data_list = []
        for k in self.records:
            data_list.append(len(self.records[k]) + 1)  # add type length
            data_list.append(k.value)
            if isinstance(self.records[k], str):
                data_list.extend([ord(c) for c in self.records[k]])

            elif isinstance(self.records[k], list):
                data_list.extend(self.records[k])

            else:
                raise NordicSemiException('Unsupported value type: 0x{:02X}'.format(type(self.records[k])))
        return data_list

    def to_c(self):
        data_list = self.to_list()
        data_len = len(data_list)
        if data_len == 0:
            return data_len, None
        else:
            self.__data_array = util.list_to_uint8_array(data_list)
            return data_len, self.__data_array.cast()

    @classmethod
    def from_c(cls, adv_report_evt):
        ad_list = util.uint8_array_to_list(adv_report_evt.data, adv_report_evt.dlen)
        ble_adv_data = cls()
        index = 0
        while index < len(ad_list):
            ad_len = ad_list[index]
            try:
                ad_type = ad_list[index + 1]
                offset = index + 2
                key = BLEAdvData.Types(ad_type)
                ble_adv_data.records[key] = ad_list[offset: offset + ad_len - 1]
            except ValueError:
                logger.error('Invalid advertising data type: 0x{:02X}'.format(ad_type))
                pass
            except IndexError:
                logger.error('Invalid advertising data: {}'.format(ad_list))
                return ble_adv_data
            index += (ad_len + 1)

        return ble_adv_data

    def __repr__(self):
        return str(self.records)


class BLEGapDataLengthParams(object):
    def __init__(self, max_tx_octets=0, max_rx_octets=0, max_tx_time_us=0, max_rx_time_us=0):
        self.max_tx_octets = max_tx_octets
        self.max_rx_octets = max_rx_octets
        self.max_tx_time_us = max_tx_time_us
        self.max_rx_time_us = max_rx_time_us

    def to_c(self):
        params = driver.ble_gap_data_length_params_t()
        params.max_tx_octets = self.max_tx_octets
        params.max_rx_octets = self.max_rx_octets
        params.max_tx_time_us = self.max_tx_time_us
        params.max_rx_time_us = self.max_rx_time_us
        return params


class BLEGapPhys(object):
    def __init__(self, tx_phys=BLEGapPhy.auto, rx_phys=BLEGapPhy.auto):
        self.tx_phys = tx_phys
        self.rx_phys = rx_phys

    def to_c(self):
        params = driver.ble_gap_phys_t()
        params.tx_phys = self.tx_phys
        params.rx_phys = self.rx_phys
        return params


"""
This example demonstrates programming with blatann in an event-driven, object-oriented manner.
The BLE device is set up in the main context, however all logic past that point is done using
event callbacks.
The main context is blocked by a "GenericWaitable", which is notified when the program
completes its intended function.

The program's logic itself is equivalent to the central example, where it connects and pairs to a device,
registers a notification callback for the counting characteristic, then tests out the conversion of strings to hex.

One thing to note: when using event-driven callbacks, it is imperative that the callbacks themselves do
not ever block on events (i.e. use the .wait() functionality). If this happens, you are essentially blocking
the event thread from processing any more events and will wait indefinitely.
A good rule of thumb when using blatann is just to not mix blocking and non-blocking calls.

This is designed to work alongside the peripheral example running on a separate nordic chip
"""
import struct
from blatann import BleDevice
from blatann.gap import smp
from blatann.examples import example_utils, constants
from blatann.waitables import GenericWaitable
from blatann.nrf import nrf_events

logger = example_utils.setup_logger(level="DEBUG")


class HexConverterTest(object):
    """
    Class to perform the hex conversion process. It is passed in the hex conversion characteristic
    and the waitable to signal when the process completes
    """
    def __init__(self, characteristic, waitable):
        """
        :param characteristic: The hex conversion characteristic to use
        :type characteristic: blatann.gatt.gattc.GattcCharacteristic
        :param waitable: The waitable to notify when the testing ends
        :type waitable: GenericWaitable
        """
        self.char = characteristic
        self.waitable = waitable
        self.i = 1
        # Generate some data, "ABCDEFG..."
        self.data_to_convert = bytes(ord('A') + i for i in range(12))

    def start(self):
        """
        Starts a new hex conversion process by writing the data to the peripheral's characteristic
        """
        data_to_send = self.data_to_convert[:self.i]
        logger.info("Converting to hex data: '{}'".format(data_to_send))
        # Write the data, and setup the _write_complete callback to be executed when the write finishes
        self.char.write(data_to_send).then(self._write_complete)

    def _write_complete(self, characteristic, event_args):
        """
        Event callback for when the hex data write completes.
        Initiates the read on the characteristic

        :param characteristic: The characteristic itself (hex converter characteristic, should always match self.char)
        :param event_args: The write event args
        :type event_args: blatann.event_args.WriteCompleteEventArgs
        """
        # At this point we should probably check event_args.status to make sure it was written successfully
        # Initiate the read, and setup _read_complete to be called when it finishes
        self.char.read().then(self._read_complete)

    def _read_complete(self, characteristic, event_args):
        """
        Event callback for when the hex characteristic read completes.
        Checks to see if the process is complete, if not starts another conversion

        :param characteristic: The characteristic itself (hex converter characteristic, should always match self.char)
        :param event_args: The read event args
        :type event_args: blatann.event_args.ReadCompleteEventArgs
        """
        logger.info("Hex: '{}'".format(event_args.value.decode("ascii")))
        self.i += 1
        if self.i > len(self.data_to_convert):
            # Done, notify the main thread that we are done
            self.waitable.notify()
        else:
            # Not done, start the next conversion
            self.start()


class MyPeripheralConnection(object):
    """
    Class to handle the post-connection database discovery and pairing process
    """
    def __init__(self, peer, waitable):
        """
        :param peer: The peer that was connected to
        :type peer: blatann.peer.Peripheral
        :param waitable: The waitable to notify to end the process
        :type waitable: GenericWaitable
        """
        self.peer = peer
        self.waitable = waitable
        self._start_db_discovery()

    def _start_db_discovery(self):
        """
        Initiates database discovery
        """
        self.peer.discover_services().then(self._on_db_discovery)

    def _on_db_discovery(self, peer, event_args):
        """
        Event callback for when database discovery completes

        :param peer: The peer the database discovery completed on (should equal self.peer)
        :type peer: blatann.peer.Peripheral
        :param event_args: The event arguments
        :type event_args: blatann.event_args.DatabaseDiscoveryCompleteEventArgs
        """

        logger.info("Service discovery complete! status: {}".format(event_args.status))
        # The peer's database is now current, log out the services found
        for service in peer.database.services:
            logger.info(service)

        # Find and subscribe to the counting characteristic
        counting_char = self.peer.database.find_characteristic(constants.COUNTING_CHAR_UUID)
        if counting_char:
            logger.info("Subscribing to the counting characteristic")
            counting_char.subscribe(self._on_counting_char_notification)
        else:
            logger.warning("Failed to find counting characteristic")

        # Initiate the pairing process
        self._start_pairing()

    def _start_pairing(self):
        """
        Sets up the security parameters and starts the pairing process
        """
        self.peer.security.set_security_params(passcode_pairing=True, io_capabilities=smp.IoCapabilities.KEYBOARD_DISPLAY,
                                               bond=False, out_of_band=False)
        self.peer.security.on_passkey_required.register(self._on_passkey_entry)
        self.peer.security.pair().then(self._on_pair_complete)

    def _on_passkey_entry(self, peer, event_args):
        """
        Event callback for when a passkey is required to be entered by the user
        Requests the user to enter the passkey and resolves the event with the passkey entered

        :param peer: the peer the passkey is for
        :param event_args: The event args
        :type event_args: blatann.event_args.PasskeyEntryEventArgs
        """
        passkey = input("Enter peripheral passkey: ")
        event_args.resolve(passkey)

    def _on_pair_complete(self, peer, event_args):
        """
        Event callback for when pairing completes. Finds the hex conversion characteristic in the database
        and starts the hex converter testing

        :param peer: The peer the pairing completed on
        :param event_args: The event args
        :type event_args: blatann.event_args.PairingCompleteEventArgs
        """
        # At this point, we should probably check event_args.status to verify that pairing completed successfully

        # Find the hex converter char. If found, start the testing. Otherwise, notify the main thread to exit
        hex_convert_char = self.peer.database.find_characteristic(constants.HEX_CONVERT_CHAR_UUID)
        if hex_convert_char:
            self.hex_counter = HexConverterTest(hex_convert_char, self.waitable)
            self.hex_counter.start()
        else:
            logger.warning("Failed to find hex convert char")
            self.waitable.notify()

    def _on_counting_char_notification(self, characteristic, event_args):
        """
            Callback for when a notification is received from the peripheral's counting characteristic.
            The peripheral will periodically notify a monotonically increasing, 4-byte integer. This callback unpacks
            the value and logs it out

            :param characteristic: The characteristic the notification was on (counting characteristic)
            :param event_args: The event arguments
            :type event_args: blatann.event_args.NotificationReceivedEventArgs
            """
        # Unpack as a little-endian, 4-byte integer
        current_count = struct.unpack("<I", event_args.value)[0]
        logger.info("Counting char notification. Current count: {}".format(current_count))


class ConnectionManager(object):
    """
    Manages scanning and connecting to the target peripheral
    """
    def __init__(self, ble_device, exit_waitable):
        """
        :param ble_device: The BLE Device to operate
        :type ble_device: BleDevice
        :param exit_waitable: The waitiable to notify to exit the program
        :type exit_waitable: GenericWaitable
        """
        self.ble_device = ble_device
        self.exit_waitable = exit_waitable
        self.target_device_name = ""
        self.connection = None
        self.peer = None

    def _on_connect(self, peer):
        """
        Event callback for when the peer's connection process completes.
        If the connection failed, will exit the program. Otherwise, will create
        a MyPeripheralConnection to manage the connection process

        :param peer: The peer that connected, or None if the connection process failed
        :type peer: blatann.peer.Peripheral
        """
        if not peer:
            logger.warning("Timed out connecting to device")
            self.exit_waitable.notify()
        else:
            logger.info("Connected, conn_handle: {}".format(peer.conn_handle))
            self.peer = peer
            # Connect the disconnect event to the exit waitable, so if
            # the peripheral disconnects from us unexpectedly the program terminates
            peer.on_disconnect.register(self.exit_waitable.notify)
            # Create the connection
            self.connection = MyPeripheralConnection(peer, self.exit_waitable)

    def _on_scan_report(self, scan_report):
        """
        Event callback when a scan report completes.
        Searches for the target advertised peripheral name. If found, starts the connection process.
        Otherwise, the program terminates

        :param scan_report: The complete scan report from the scanning session
        :type scan_report: blatann.gap.scanning.ScanReportCollection
        """
        for report in scan_report.advertising_peers_found:
            if report.advertise_data.local_name == self.target_device_name:
                logger.info("Found match: connecting to address {}".format(report.peer_address))
                self.ble_device.connect(report.peer_address).then(self._on_connect)
                return

        logger.info("Did not find target peripheral")
        self.exit_waitable.notify()

    def scan_and_connect(self, name, timeout=4):
        """
        Starts the scanning process and sets up the callback for when scanning completes

        :param name: The name of the peripheral to look for
        :param timeout: How long to scan for
        """
        logger.info("Scanning for '{}'".format(name))
        self.target_device_name = name
        self.ble_device.scanner.set_default_scan_params(timeout_seconds=timeout)
        self.ble_device.scanner.start_scan().then(self._on_scan_report)


def main(serial_port):
    target_device_name = constants.PERIPHERAL_NAME
    # Create a waitable that will block the main thread until notified by one of the classes above
    main_thread_waitable = GenericWaitable()

    # Create and open the BLE device (and suppress spammy logs)
    ble_device = BleDevice(serial_port)
    ble_device.event_logger.suppress(nrf_events.GapEvtAdvReport)
    ble_device.open()

    # Create the connection manager and start the scanning process
    s = ConnectionManager(ble_device, main_thread_waitable)
    s.scan_and_connect(target_device_name)

    # Block the main thread indefinitely until the program finishes
    main_thread_waitable.wait()

    # Clean up, disconnect if a peer was connected to
    if s.peer:
        s.peer.disconnect().wait()
    ble_device.close()


if __name__ == '__main__':
    main("COM9")
